import pickle
from Bio.Seq import Seq
import os
import pandas as pd
import numpy as np
from ExpoSeq.augment_data.binding_data import collect_binding_data
from Bio.Seq import Seq



class SequencingReport:
    def __init__(self,sequencing_report):
        self.origin_seq_report = sequencing_report.copy()
        self.sequencing_report = sequencing_report

    def get_exp_names(self, experiment_path):
        try:
            with open(experiment_path, "rb") as f:
                unique_experiments = pickle.load(f)
        except:
            unique_experiments = self.sequencing_report["Experiment"].unique().tolist()
            unique_experiments = dict(zip(unique_experiments, unique_experiments))
        return unique_experiments
            
    def map_exp_names(self, unique_experiments):
        self.sequencing_report["Experiment"] = self.sequencing_report["Experiment"].map(unique_experiments)


    def is_divisible_by_three(self, seq):
        return len(seq) % 3 == 0

    @staticmethod
    def translate_seq(nucleotide):
        return Seq(nucleotide).translate()

    def get_fragment(self):
        threshold_count = 0.1 * len(self.origin_seq_report) # to get sequences which have values
        return [col for col in self.origin_seq_report.columns if col.startswith('nSeq') and self.origin_seq_report[col].count() > threshold_count]

    
    def filter_region(self, region_string):
        added_columns = ["nSeq" + region_string, "aaSeq" + region_string] #"minQual" + region_string,
        fixed_cols = ["Experiment", "cloneId", "readCount", "readFraction"]
        cols_of_interest = fixed_cols + added_columns
        self.sequencing_report = self.sequencing_report[cols_of_interest]
        self.sequencing_report = self.sequencing_report[self.sequencing_report['nSeq' + region_string].apply(self.is_divisible_by_three)]
       # self.sequencing_report["aaSeq" + region_string] = self.sequencing_report["nSeq" + region_string].apply(self.translate_sequence)
    
        

    def trim_data(self, region_string, divisible_by = 3,length_threshold = 9, min_read_count = 0, new_fraction = "cloneFraction"):

        sequencing_report = self.sequencing_report.groupby("Experiment").apply(lambda group: group.drop_duplicates(subset=["nSeq" + region_string], keep="first")).reset_index(drop=True)
        indexes_to_drop = sequencing_report[sequencing_report["aaSeq" + region_string] == 'region_not_covered'].index
        sequencing_report = sequencing_report.drop(indexes_to_drop)

        sequencing_report["lengthOfCDR3"] = sequencing_report["nSeqCDR3"].str.len() ## assumes all that pipeline starts with cdr3 region

        sequencing_report = sequencing_report.reset_index()

        sequencing_report = sequencing_report[(sequencing_report["lengthOfCDR3"] % divisible_by) == 0]
        sequencing_report = sequencing_report[(sequencing_report["aaSeq" + region_string].str.len()) > length_threshold ]
        sequencing_report = sequencing_report[(sequencing_report["readCount"] > min_read_count)]


        new_column = sequencing_report['readCount'] / sequencing_report.groupby('Experiment')['readCount'].transform('sum')
        self.sequencing_report = sequencing_report.copy()
        self.sequencing_report[new_fraction] = np.array(new_column)
        
    def remove_not_covered(self):
        self.sequencing_report = self.sequencing_report[~self.sequencing_report.applymap(lambda x: x == "region_not_covered").any(axis=1)]
    
    def prepare_seq_report(self, region_string, divisible_by, length_threshold, min_read_count):
        self.filter_region(region_string)
        self.trim_data(region_string, divisible_by, length_threshold, min_read_count,)
        self.remove_not_covered()
        
    def find_longest_sequence(self):
        columns = self.origin_seq_report.columns.to_list()
        regions_order = ['nSeqFR1', 'nSeqCDR1', 'nSeqFR2', 'nSeqCDR2', 'nSeqFR3', 'nSeqCDR3', 'nSeqFR4']
        # Map regions to their indices in the order
        indices = [regions_order.index(col) for col in columns if col in regions_order]
        # If no regions are present, return an empty list
        if not indices:
            return []
        # Find the longest consecutive subsequence
        longest_seq = []
        current_seq = [indices[0]]
        for i in range(1, len(indices)):
            if indices[i] - current_seq[-1] == 1:  # Check for consecutiveness
                current_seq.append(indices[i])
            else:
                # If not consecutive, update longest sequence if needed and reset current sequence
                if len(current_seq) > len(longest_seq):
                    longest_seq = current_seq
                current_seq = [indices[i]]
        # Handle the case where the longest sequence is at the end
        if len(current_seq) > len(longest_seq):
            longest_seq = current_seq
        # Convert indices back to region names
        return [regions_order[idx] for idx in longest_seq]

        
    def filter_longest_sequence(self, divisible_by= 3, length_threshold= 9, min_read_count = 0):
        cols_longest = self.find_longest_sequence()
        region_string = cols_longest[0]
        for i in cols_longest:
            i = i.replace("nSeq", "")
            self.trim_data(i, divisible_by= divisible_by, length_threshold= length_threshold, min_read_count = min_read_count)
        cols_of_interest = ["Experiment", "cloneId", "readCount", "cloneFraction"] + cols_longest
        found_frags = self.sequencing_report[cols_of_interest]
        found_frags = found_frags[~found_frags.applymap(lambda x: x == "region_not_covered").any(axis=1)]
        found_frags['merged_with_spaces'] = found_frags[cols_longest].apply(lambda row: ' '.join(row), axis=1)
        found_frags["nSeqall"] = found_frags['merged_with_spaces'].str.replace('\\', '').str.replace(' ', '')
        found_frags['aaSeqall'] = found_frags['nSeqall'].apply(self.translate_seq)
        found_frags.reset_index(inplace = True)
        region_string = "all"
        self.sequencing_report = found_frags
        return region_string

        



class BindingReport:
    def __init__(self, module_dir, experiment):
        self.binding_data_dir = os.path.join(module_dir,
                                "my_experiments",
                                experiment,
                                "binding_data.csv")
        
    def ask_binding_data(self):
        if not os.path.isfile(self.binding_data_dir):
            add_binding = input("Do you have binding Data? Y/n")
            if add_binding.lower() in ["Y", "y"]:
                binding_data = collect_binding_data()

                binding_data.to_csv("binding_data.csv")
            else:
                binding_data = None
        else:
            binding_data = pd.read_csv(self.binding_data_dir)
        return binding_data








import pandas as pd
import numpy as np
def calculate_entropy(probs):
    """Calculate Shannon entropy."""
    return -np.sum([p * np.log2(p) if p > 0 else 0 for p in probs])
def calculate_bit_values(entropies):
  """Calculates the bit values for a sequence logo, given a list of Shannon entropies."""

  bit_values = []
  for entropy in entropies:
    bit_value = 2 - entropy
    if bit_value < 0:
      bit_value = 0
    bit_values.append(bit_value)
  return bit_values
def cleaning(sample_name, report, chosen_seq_length, region_string, method):
    sample = report[report["Experiment"] == sample_name]
    local_report = sample[["Experiment", "cloneFraction", region_string]]
    sequences = local_report[region_string]
    #local_report = local_report.sort_values("cloneFraction").groupby("Experiment", as_index = False).head(3) # filters out three highest values of each group
    aminoacids = "ACDEFGHIKLMNPQRSTVWY"

    compDict = {aa: chosen_seq_length*[0] for aa in aminoacids}
    sequences = local_report[local_report.aaSeqCDR3.astype(str).str.len() == chosen_seq_length]["aaSeqCDR3"]
    length_filtered_seqs = sequences.shape[0]
    for seq in sequences:
        for aa_position in range(len(seq)):
            aminoacid = seq[aa_position]
            if aminoacid == '*':
                pass
            else:
                compDict[aminoacid][aa_position] += 1
    if method == "bits":
        frequencies = {aa: [count / length_filtered_seqs for count in compDict[aa]] for aa in aminoacids}
        # Calculate Shannon entropy for each position
        entropies = [calculate_entropy([frequencies[aa][i] for aa in aminoacids]) for i in range(chosen_seq_length)]
        # Calculate bits for sequence logo for each amino acid
        bits_dict = {aa: [2 - entropies[i] if frequencies[aa][i] > 0 else 0 for i in range(chosen_seq_length)] for aa in aminoacids}
        aa_distribution = pd.DataFrame.from_dict(bits_dict)
    else:
        aa_distribution = pd.DataFrame.from_dict(compDict)
        aa_distribution = aa_distribution.divide(aa_distribution.sum(axis=1), axis=0)
    return aa_distribution, chosen_seq_length, length_filtered_seqs





import os.path
from os import path

from glob import glob
try:
    import tkinter as tk
    from tkinter import filedialog
except:
    pass
def collectFiles(filetypes = ('.txt', ".tsv", '.fastq_AlignmentReport')):
    filenames = []
    try:
        data_folder = filedialog.askdirectory()
    except:
        while True:
            data_folder = input("copy and paste the path to the directory without any / as ending.")
            if path.isdir(os.path.abspath(data_folder)) == True:
                break
    data_folder = path.abspath(data_folder) + '\\*'
        # maybe create user input who can choose which filetypes he/she wants to include
    for files in filetypes:
        filenames.extend(glob(data_folder+files)) # might not be the same name for everyone, has to be changed if package should be used by everyone
    filenames.sort(key = len)
    return filenames








import pandas as pd
from .bool_sequences_matrix import find_seq_matches
import numpy as np

#sequencing_report = pd.read_table(r"C:\Users\nilsh\PycharmProjects\ExpoSeq\try_outs\test_data\sequencing_report.txt", sep = ",")
#sequencing_report = sequencing_report.drop('Unnamed: 0', axis=1)
#protein = True
def cleaning_jaccard(sequencing_report, region_of_interest, protein, specific_experiments):
    heatmap_axis, unique_sequences, unique_experiments = find_seq_matches(sequencing_report,
                                                                          region_of_interest,
                                                                          protein,
                                                                          specific_experiments)
    heatmap_absolute_jaccard = np.zeros([heatmap_axis, heatmap_axis])
    for index_sample_one in range(heatmap_axis):
        d1 = unique_sequences.iloc[:, index_sample_one]
        for index_sample_two in range(heatmap_axis):
            d2 = unique_sequences.iloc[:, index_sample_two]
            lib_matches = d1 * d2
            lib_matches = lib_matches.sum()
            a = lib_matches
            b = d1.sum() - lib_matches
            c = d2.sum() - lib_matches
            heatmap_absolute_jaccard[index_sample_one, index_sample_two] = lib_matches/(lib_matches+b+c)
            heatmap_absolute_jaccard[index_sample_two, index_sample_one] = lib_matches / (lib_matches + b + c)
    matrix = pd.DataFrame(heatmap_absolute_jaccard,
                          index = list(unique_experiments),
                          columns = list(unique_experiments))

    return matrix, unique_sequences, unique_experiments



  #  if rename_from_dic == True:
   #     matrix.rename(columns = Library_2_to_panning,
    #                  inplace = True)
     #   matrix.rename(index = Library_2_to_panning,
      #                inplace = True)


import re
import numpy as np
import pandas as pd



class read_alignment_report:
    def __init__(self, alignment_report):
        self.report = alignment_report
        self.digit = "\d+"

    def find_word_align_rep(self, word, regex): #"Total sequencing reads"  # auomate this or ask user how it is called in his / her file
        for line in self.report:
            if line.find(word) != -1:
                preferred_line = line
                total_sequenced_reads = re.search(regex,
                                                preferred_line).group()
        return total_sequenced_reads






from ExpoSeq.tidy_data.tidy_stacked_aa_distr import cleaning
import matplotlib.pyplot as plt

def stacked_aa_distr(ax, sequencing_report, sample, region, protein, font_settings,region_string):

    aa_distribution = cleaning(sequencing_report, sample, region, region_string)
    color_scheme = {
        'F': [.16, .99, .18],
        'Y': [.04, .40, .05],
        'L': [.99, .60, .25],
        'V': [1.0, .80, .27],
        'I': [.80, .60, .24],
        'H': [.40, .02, .20],
        'W': [.42, .79, .42],
        'A': [.99, .60, .42],
        'S': [.04, .14, .98],
        'T': [.17, 1.0, 1.0],
        'M': [.80, .60, .80],
        'N': [.21, .40, .40],
        'Q': [.40, .41, .79],
        'R': [.59, .02, .04],
        'K': [.40, .20, .03],
        'E': [.79, .04, .22],
        'G': [.95, .94, .22],
        'D': [.99, .05, .11],
        'P': [.10, .61, .99],
        'C': [.09, .60, .60],
    }
    color_list = [color_scheme[aa] for aa in aa_distribution.columns]
    aa_distribution.plot(kind='bar', stacked=True,color=color_list, ax = ax)
    ax.set_xlabel('Position on amino acid sequence', **font_settings)
    if protein == True:
        ax.set_ylabel('Relatvie Proportion of Amino Acid', **font_settings)
    else:
        ax.set_ylabel('Relatvie Proportion of Nucleotide', **font_settings)

#    ax.set_xticks(rotation = 360)
    ax.legend(bbox_to_anchor=(1.05, 1), loc='upper left')
    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 22
    if protein == True:
        ax.set_title("Distribution of Aminoacids per Position for your sequences",pad = 12, **font_settings)
    else:
        ax.set_title("Distribution of Nucleotides per Position for your sequences",pad = 12, **font_settings)
    ax.set_xticklabels([*range(region[0], region[1]+1)], rotation=0, ha='center')
    font_settings["fontsize"] = original_fontsize


import numpy as np
import pandas as pd


def cleaning_data(sequencing_report, region_of_interest, protein = True, specific_experiments = False):
 #   new_fraction = 'new_fraction'
    if specific_experiments != False:
        sequencing_report = sequencing_report[sequencing_report['Experiment'].isin(specific_experiments)]
    else:
        pass
   # new_column = sequencing_report['readCount'] / sequencing_report.groupby('Experiment')['readCount'].transform('sum')
    #sequencing_report[new_fraction] = np.array(new_column)
    if protein == True:
   #     sequencing_report = mapFunc(sequencing_report = sequencing_report,
    #                                column = 'nSeqCDR3',
     #                               func = genetic_dogma,
       #                             column_name = 'peptide_seq')
        strand_column =  region_of_interest
    else:
        strand_column = region_of_interest


    unique_experiments = sequencing_report["Experiment"].unique()
    group_columns = ["Experiment", strand_column]
   # column = sequencing_report.groupby(group_columns)[new_fraction].transform('sum')
    #sequencing_report[new_fraction] = column

   # new_column = sequencing_report['readCount'] / sequencing_report.groupby('Experiment')['readCount'].transform('sum')
   # sequencing_report = sequencing_report.copy()
    #sequencing_report["cloneFraction"] = new_column


    unique_sequences = pd.DataFrame(sequencing_report[strand_column].unique()) #nseqcdr3 has to be changed since it is a chosen name for column
    unique_sequences.rename(columns={0: strand_column},
                            inplace=True)
    unique_experiments = list(unique_experiments)

    for experiment in unique_experiments:
        local_data = sequencing_report[sequencing_report["Experiment"] == experiment][[strand_column,
                                                                                       "cloneFraction"]]

        unique_sequences = unique_sequences.merge(local_data,
                                                  how='left',
                                                  on = strand_column)
       # unique_sequences = unique_sequences.fillna(0)
        unique_sequences = unique_sequences.rename(columns={"cloneFraction": "cloneFraction" + experiment})
    unique_sequences.drop(strand_column,
                         inplace=True,
                          axis=1)
    unique_sequences = unique_sequences.fillna(0)
    unique_sequences.reset_index(inplace = True)
    return unique_sequences, unique_experiments



def morosita_horn_matrix(unique_sequences, unique_experiments):

    columns = len(unique_experiments)
    morosita_horn_matrix = np.zeros((columns,
                                     columns))
    for index_sample_one in range(columns):
        d1 = unique_sequences.iloc[:, index_sample_one]
        start = 0
        for index_sample_two in range(columns):
            d2 = unique_sequences.iloc[:, index_sample_two]
            product_d1_d2 = np.array(d1) * np.array(d2)
            d1_square = np.array(d1)**2
            d2_square = np.array(d2)**2
       #     cal_cols=pd.concat([d1,d2,product_d1_d2,d1_square,d2_square], axis=1).fillna(0)
        #    cal_cols.columns=['d1',
         #                     'd2',
          #                    'product_d1_d2',
           #                   'd1_square',
            #                  'd2_square']
            #sub = cal_cols['product_d1_d2'] > 0
            #subset = cal_cols[sub]
            MH = (2*np.sum(product_d1_d2)) /(np.sum(d1_square) + np.sum(d2_square))
            morosita_horn_matrix[index_sample_one, index_sample_two] = MH
            morosita_horn_matrix[index_sample_two, index_sample_one] = MH
          #  MH = 2*subset['product_d1_d2'].sum()/(subset['d1_square'].sum()+subset['d2_square'].sum())
           # morosita_horn_matrix[index_sample_one, index_sample_two] = MH
            #morosita_horn_matrix[index_sample_two, index_sample_one] = MH
            start += 1
    matrix = pd.DataFrame(morosita_horn_matrix,
                          index = list(unique_experiments),
                          columns = list(unique_experiments))
    return matrix, unique_sequences, unique_experiments






from glob import glob
import subprocess
from ast import literal_eval
import pickle
import pandas as pd
from ExpoSeq.augment_data.trimming import trimming
import pkg_resources
import os

#psutil
try:
    import tkinter as tk
    from tkinter import filedialog
except:
    pass
from ExpoSeq.augment_data.create_mixcr_commands import CreateCommand
from ExpoSeq.augment_data.collect_fastqs import CollectFastq
from ExpoSeq.settings.check_dirs import check_dirs

def check_mixcr(path_to_mixcr, data, settings_dir, testing = False):
    if path_to_mixcr == "":
        while True:
            print("choose the mixcr java file with the file chooser")
            try:
                path_to_mixcr = filedialog.askopenfilename()
            except:
                path_to_mixcr = input("copy and paste the path to the mixcr jar file here.")
                path_to_mixcr = os.path.normpath(path_to_mixcr)
            if testing == True:
                break

            try:
                print("Test Mixcr")
                subprocess.run(["java",
                                "-jar",
                                path_to_mixcr,
                                "--version"
                                ])
                confirmation = True
            except:
                print("Apparently you have not choosen the right path to the mixcr .jar file. Please try again")
                confirmation = False
            if confirmation == True:
                path_to_mixcr = os.path.abspath(path_to_mixcr)
                data["mixcr_path"] = path_to_mixcr
                with open(settings_dir, "w") as f:
                    f.write(str(data))
                break
    else:
        pass

def process_mixcr(experiment, method, testing, paired_end_sequencing):
    pkg_path = pkg_resources.resource_filename("ExpoSeq", "")
    module_dir = os.path.abspath("")
    check_dirs(module_dir, experiment)
    if not testing:
        Collect = CollectFastq(paired_end_sequencing)
        Collect.get_files()
        files = Collect.paired
        files = [[i] for i in files]
        if paired_end_sequencing:
            while True:
                if len(files) > 0:
                    Collect.call_pairs()
                    correct_matches = input("Are the matches of forward and backward reads correct? Y/n")
                else:
                    correct_matches = "n"
                    print("You will need to restart the collection of the files if you want to proceed with paired end sequencing data.\nMake sure that you have the forward and backward files in two seperate folder and that you have the same number of files.")
                if correct_matches in ["Y", "y", "N", "n"]:
                    if correct_matches in ["Y", "y"]:
                        break
                    else:
                        while True:
                            choose_strat = input("You can restart the collection of the files or manually add them. \nPress 1 for restarting the collection by choosing the directories.\nPress 2 for restarting the collection and manually choose the forward and backward reads.")
                            if choose_strat in ["1", "2"]:
                                Collect = CollectFastq(paired_end_sequencing)
                                if choose_strat == "1":
                                    Collect.get_files()
                                    files = Collect.paired
                                    break
                                else:
                                    Collect.manually_match_pairs()
                                    files = Collect.paired
                                    break
                            else:
                                print("Please enter a valid value")
                else:
                    print("Please enter a valid value")
        else:
            pass

    else:
        files = glob(os.path.join(pkg_path, "test_data", "test_files", "*.fastq"))
        print(files)
        files = [[i] for i in files]
        experiment = "test_directory"
    settings_dir = os.path.join(module_dir,
                                "settings",
                                "global_vars.txt")
    with open(settings_dir) as f:
        data = f.read()
    data = literal_eval(data)
    path_to_mixcr = data["mixcr_path"]
    check_mixcr(path_to_mixcr, data, settings_dir)

    sequencing_report = pd.DataFrame([])
    while True:
        java_heap_size = input(
            "Enter the amount of memory you want to allocate for java in Mb. Normally 1000 is sufficient.")
        try:
            java_heap_size = int(java_heap_size)
            break
        except:
            print("Please enter the amount as integer")

    for filename in files:

        for file in os.listdir(os.path.join(module_dir, "temp")):
            os.remove(os.path.join(module_dir,
                                   "temp",
                                   file))
        Commands = CreateCommand(module_dir, path_to_mixcr, paired_end_sequencing, experiment, filename, java_heap_size)
        subprocess.run(Commands.prepare_align(method))
        skip_sample = False
        try:
            align_quality, percent_quality = Commands.read_aligned_reads()
            if align_quality < 1000 and align_quality >10:
                print(f"Only {align_quality} reads could be aligned ({percent_quality}).")
            if align_quality < 10:
                print("Too less reads were aligned. Please choose the correct mixcr method.")
                skip_sample = True
            if skip_sample == False:
                subprocess.run(Commands.prepare_assembly())
                assembly_qual, realtive_assembly_quality = Commands.read_clones()
                if assembly_qual < 1000 and realtive_assembly_quality > 10:
                    print(f"Only {assembly_qual} reads could be assembled ({realtive_assembly_quality}).")
                if assembly_qual < 10:
                    print("Too less reads were assembled. Please choose the correct ")
                    skip_sample = True
                if skip_sample == False:
                    subprocess.run(Commands.prepare_clones())
                    basename = Commands.basename
                    clones_sample = pd.read_table(Commands.table_tsv)
                    clones_sample["Experiment"] = basename
                    sequencing_report = pd.concat([sequencing_report, clones_sample])

                    for file in os.listdir(os.path.join(module_dir, "temp")):
                        os.remove(os.path.join(module_dir,
                                            "temp",
                                            file))
        except:
            print(f"File {filename} could not be processed. Please check if you have chosen the correct mixcr method or verify that the sequencing was successful.")


    if not testing:
        
        if sequencing_report.shape[0] == 0:
            raise Exception(
                "The processing failed for all of your files. Most likely because you chose the wrong mixcr method.")
        report_dir = os.path.join(module_dir,
                                  "my_experiments",
                                  experiment,
                                  "sequencing_report.csv")
        sequencing_report.to_csv(report_dir)
        data["last_experiment"] = experiment
        glob_vars_dir = os.path.join(module_dir,
                                     "settings",
                                     "global_vars.txt")
        with open(glob_vars_dir, "w") as f:
            f.write(str(data))
        unique_experiments = sequencing_report["Experiment"].unique()
        experiment_dic = {item: item for item in list(unique_experiments)}
        exp_names_dir = os.path.join(module_dir,
                                     "my_experiments",
                                     experiment,
                                     "experiment_names.pickle")
        with open(exp_names_dir, "wb") as f:
            pickle.dump(experiment_dic, f)
    print(
        "The output of your processing is a sequencing_report which contains the sequences and their sample names of all fastq files. Further, a folder with the alignment reports for each fastq file was created. You can have a look under ~/my_experiments/YOUR_EXPERIMENT_NAME")





from pandas import DataFrame

def cleaning_data(all_alignment_reports, sequencing_report_all):
    all_alignment_reports = all_alignment_reports.sort_values("Input file(s)").reset_index(drop=True)
    unique_experiments = sequencing_report_all.sort_values("Experiment")["Experiment"].unique()
    unique_experiments = list(unique_experiments)
    tot_sequencing_reads = all_alignment_reports["Total sequencing reads"]
    aligned_reads = all_alignment_reports["Successfully aligned reads"].str.split("(", expand = True).iloc[:, 0]
    data = DataFrame([unique_experiments, aligned_reads, tot_sequencing_reads]).T
    data.columns = ["Experiment", "Aligned_Reads", "tot_sequenced_reads"]
    return data

###### old version
#def cleaning_data(common_vars, local_pattern_more_digits):
 #   grouped_filenames = sequencing_report_all.unique()
  #  boxplot_data_frame = pd.DataFrame()
   # report_name = "fastq_AlignmentReport"
    #max_files = len(grouped_filenames)
    #limit = int(input(f"how many do you want to plot? Max No. is {max_files}"))
    #current_status = 0
    # # not in class because it reduces flexibility
    #not_used = []
    #for cluster in grouped_filenames:
     #   if current_status == limit:
      #      break
       # else:
        #    for file in cluster:
         #       filename = re.search(local_pattern_more_digits,
          #                          file).group()
           #     if report_name in file:
            #        alignment_frame = collect_boxplot_data(file,
             #                                              local_pattern_more_digits)
              #      boxplot_data_frame = pd.concat([boxplot_data_frame, alignment_frame],
               #                                     ignore_index = False)
                #    file_found = True
                 #   current_status += 1
                  #  break
              #  else:
               #     if file == cluster[len(cluster)-1]:
                #        not_used.append(filename)
                 #   pass
   # experiment = boxplot_data_frame["Experiment"].map(Library_2_to_panning)
    #boxplot_data_frame["Experiment"] = experiment
    #return boxplot_data_frame



import pandas as pd
from ExpoSeq.tidy_data.tidy_cluster_embedding import tidy_embed

def show_difference(sequencing_report, list_experiments,strands, batch_size, pca_components, perplexity, iterations_tsne,region_string, ax, legend_settings, font_settings):
    tsne_results, aminoacids, experiments_batch = tidy_embed(sequencing_report,
                                                              batch_size,
                                                              list_experiments,
                                                              pca_components,
                                                              perplexity,
                                                              iterations_tsne,
                                                              region_string)
    tsne_plot = ax.scatter(tsne_results.tsne1,
                           tsne_results.tsne2,
                           c = pd.factorize(experiments_batch)[0],
                            )
    ax.legend(handles=tsne_plot.legend_elements()[0],
               labels=list_experiments,
              **legend_settings)
    x = tsne_results["tsne1"].values.tolist()
    y = tsne_results["tsne2"].values.tolist()
    ax.set_xlabel("t-SNE1", **font_settings)
    ax.set_ylabel("t-SNE2", **font_settings)
    if strands == True:
        for i in range(0, len(x), 10):
            ax.annotate(aminoacids[i],
                        (x[i][0], y[i][0]),
                        fontsize = 5,
                        )
    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 22
    ax.set_title("Sequence Embedding for given Samples", pad= 12, **font_settings)
    font_settings["fontsize"] = original_fontsize


from .plots import barplot, cluster_embedding, embedding_with_binding, hist_lvst_dist, length_distribution, \
    logo_plot, plt_heatmap, relative_sequence_abundance, stacked_aa_distribution, usq_plot, levenshtein_clustering
import matplotlib.pyplot as plt
from .augment_data.binding_data import collect_binding_data
from .augment_data.uploader import upload
from ast import literal_eval
import pandas as pd
import pickle
import os
from ExpoSeq.augment_data.randomizer import create_sequencing_report, create_binding_report
from .design.multi_seq_align import MSA
from ExpoSeq.settings import change_settings, change_save_settings, reports, plot_styler
from tkinter import filedialog
import subprocess
import glob
from ExpoSeq.augment_data.uploader import create_alignment_report
from ExpoSeq.settings.general_instructions import print_instructions
from ExpoSeq.tidy_data.heatmaps.read_matrix import read_matrix
from ExpoSeq.augment_data.mixcr_nils import check_mixcr

class MyFigure:
    def __init__(self):
        self.fig = plt.figure(1, figsize = (12, 10))
        self.ax = self.fig.gca()
        self.plot_type = "multi"

    def check_fig(self, ):
        if not plt.fignum_exists(1):
            self.fig = plt.figure(1)

    def clear_fig(self):
        self.fig.clear()
        if self.plot_type == "single":
            self.ax = self.fig.gca()

    def update_plot(self):
        self.ax = self.fig.gca()
        self.style = plot_styler.PlotStyle(self.ax, self.plot_type)
       # figManager = plt.get_current_fig_manager()
      #  figManager.window.showMaximized()
        self.tighten()

    def tighten(self):
        plt.tight_layout()


def save_matrix(matrix, path = None):
    if path == None:
        while True:
            save_matrix = input("Do you want to save the generated data? (Y/n)")
            if save_matrix in ["Y", "y", "N", "n"]:
                break
            else:
                print("Please enter Y or n")
        if save_matrix.lower() in ["Y", "y"]:
            while True:
                filename_matrix = input("Enter a name for the file. The file will be saved locally in your IDE.")
                if not os.path.isfile(filename_matrix):
                    matrix.to_excel(path + ".xlsx")
                    break
                else:
                    print("This file already exists. Please choose another name.")
    else:
        matrix.to_excel(path)



class Directories:
    def __init__(self):
        self.module_dir = os.path.abspath("")
        self.common_vars = os.path.join(self.module_dir, "settings", "global_vars.txt")
        self.font_settings_path = os.path.join(self.module_dir, "settings", "font_settings.txt")
        self.legend_settings_path = os.path.join(self.module_dir,
                                                 "settings",
                                                 "legend_settings.txt")
        self.colorbar_path = os.path.join(self.module_dir,
                                          "settings",
                                          "colorbar.txt")

    def check_dirs(self):
        if not os.path.isdir(os.path.join(self.module_dir, "my_experiments")):
            print("create my_experiments directory")
            os.mkdir(os.path.join(self.module_dir, "my_experiments"))
        if not os.path.isdir(os.path.join(self.module_dir, "temp")):
            print("create temp directory")
            os.mkdir(os.path.join(self.module_dir, "temp"))
        if not os.path.isdir(os.path.join(self.module_dir, "settings")):
            print("create settings directory")
            os.mkdir(os.path.join(self.module_dir, "settings"))

    def create_global_params(self):
        common_vars = {'mixcr_path': '', 'last_experiment': '', 'api_gpt3': '', 'region_of_interest': '', 'RAM': '',
                       'clustalw_path': ''}
        with open(self.common_vars, "w") as f:
            f.write(str(common_vars))

    def create_font_settings(self):
        font_settings = {'fontfamily': 'serif', 'fontsize': '18', 'fontstyle': 'normal', 'fontweight': 'bold'}
        with open(os.path.join(self.module_dir, "settings", "font_settings.txt"), "w") as f:
            f.write(str(font_settings))

    def create_legend_settings(self):
        legend_settings = {'loc': 'upper right', 'bbox_to_anchor': (1, 1), 'ncols': 1, 'fontsize': 16, 'frameon': True,
                           'framealpha': 1, 'facecolor': 'white', 'mode': None, 'title_fontsize': 'small'}
        with open(os.path.join(self.module_dir, "settings", "legend_settings.txt"), "w") as f:
            f.write(str(legend_settings))

    def create_colorbar_settings(self):
        colorbar = {'cmap': 'inferno', 'orientation': 'vertical', 'spacing': 'proportional', 'extend': 'neither'}
        with open(self.colorbar_path, "w") as f:
            f.write(str(colorbar))

    def read_global_params(self):
        if not os.path.isfile(self.common_vars):
            self.create_global_params()
        with open(self.common_vars, "r") as f:
            global_params = f.read()
        global_params = literal_eval(global_params)
        return global_params

    def read_font_settings(self):
        if not os.path.isfile(self.font_settings_path):
            self.create_font_settings()
        with open(self.font_settings_path, "r") as f:
            font_settings = f.read()
        font_settings = literal_eval(font_settings)
        return font_settings

    def read_legend_settings(self):
        if not os.path.isfile(self.legend_settings_path):
            self.create_legend_settings()
        with open(self.legend_settings_path, "r") as f:
            legend_settings = f.read()
        legend_settings = literal_eval(legend_settings)
        return legend_settings

    def read_colorbar_settings(self):
        if not os.path.isfile(self.colorbar_path):
            self.create_colorbar_settings()
        with open(self.colorbar_path, "r") as f:
            colorbar_settings = f.read()
        colorbar_settings = literal_eval(colorbar_settings)
        return colorbar_settings

    def get_experiment_path(self, experiment):
        experiment_path = os.path.join(self.module_dir,
                                       "my_experiments",
                                       experiment,
                                       "experiment_names.pickle")
        return experiment_path





# Call the function to print the instructions

class PlotManager:
    def __init__(self,experiment = None, test_version=False, test_exp_num=3, test_panrou_num=1, divisible_by=3, length_threshold=9,
                 min_read_count=3):
        self.is_test = test_version
        self.Settings = change_settings.Settings()
        self.Settings.check_dirs()
        self.global_params = self.Settings.read_global_vars()
        self.module_dir = self.Settings.module_dir
        if test_version == True:
            self.experiment = "Test"
            self.sequencing_report = create_sequencing_report(num_experiments=test_exp_num,
                                                              panning_rounds=test_panrou_num,
                                                              )
            self.alignment_report = None
            self.binding_data = create_binding_report(self.sequencing_report,
                                                      num_antigen=test_panrou_num)
            self.region_string = "CDR3"
        else:
            if experiment == None:
                self.sequencing_report, self.alignment_report, self.experiment = upload()
            else:
                self.sequencing_report  = pd.read_csv(os.path.join(self.module_dir,
                                                                   "my_experiments",
                                                                   experiment,
                                                                   "sequencing_report.csv"))
                self.alignment_report = create_alignment_report(self.module_dir, experiment)
                self.experiment = experiment
            self.region_string = self.global_params["region_of_interest"]
            if self.region_string == '':
                self.region_string = "CDR3"
            binding_report = reports.BindingReport(self.module_dir,
                                                   self.experiment)
            self.binding_data = binding_report.ask_binding_data()
        self.Report = reports.SequencingReport(self.sequencing_report)
        experiment_path = self.Settings.get_experiment_path(self.experiment)
        self.unique_experiments = self.Report.get_exp_names(experiment_path)
        self.Report.prepare_seq_report(self.region_string,
                                       divisible_by=divisible_by,
                                       length_threshold=length_threshold,
                                       min_read_count=min_read_count)
        self.Report.map_exp_names(self.unique_experiments)
        self.avail_regions = self.Report.get_fragment()
        self.sequencing_report = self.Report.sequencing_report
        self.font_settings = self.Settings.read_font_settings()
        self.legend_settings = self.Settings.read_legend_settings()
        self.colorbar_settings = self.Settings.read_colorbar_settings()
        self.ControlFigure = MyFigure()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax, self.ControlFigure.plot_type)
        # self.settings_saver = change_save_settings.Change_save_settings()
        self.experiments_list = list(self.unique_experiments.values())
        self.region_of_interest = "aaSeq" + self.region_string
        self.preferred_sample = self.experiments_list[0]
        self.change_preferred_antigen()
        print_instructions()
        self.report_path = os.path.join(self.module_dir, "my_experiments", self.experiment, "reports_pipeline")
        if not os.path.isdir(self.report_path):
            os.mkdir(self.report_path)
        if not os.path.isdir(os.path.join(self.module_dir, "my_experiments", self.experiment, "plots")):
            os.mkdir(os.path.join(self.module_dir, "my_experiments", self.experiment, "plots"))
        self.plot_path = os.path.join(self.module_dir, "my_experiments", self.experiment, "plots", self.region_of_interest)
        self.mixcr_plots_path = os.path.join(self.plot_path, "mixcr_plots")
        if not os.path.isdir(self.plot_path):
            os.mkdir(self.plot_path)
            if not os.path.isdir(self.mixcr_plots_path):
                os.mkdir(self.mixcr_plots_path)
            self.full_analysis()


    def save_in_plots(self, enter_filename):
        plt.savefig(fname = os.path.join(self.plot_path, enter_filename + f".png"), dpi = 300,  format="png")

    def get_best_binder(self):
        if self.binding_data is None:
            return None
        found_antigens = self.binding_data.columns.to_list()[1:]
        best_binding = {}

        for i in self.unique_experiments:
            sub_report = self.sequencing_report[self.sequencing_report["Experiment"] == i]
            merged_report = sub_report.merge(self.binding_data, how = "inner")
            clone_fractions = merged_report["cloneFraction"]
            clone_fractions.fillna(0, inplace = True)
            best_binding_local = {}
            for antigen in found_antigens:
                binding_values = merged_report[antigen]
                binding_values.fillna(0, inplace = True)
                significance_binding = (binding_values * clone_fractions).mean()
                if antigen in best_binding_local.keys():
                    if best_binding_local.get(antigen) < significance_binding:
                        best_binding_local[antigen] = significance_binding
                    else:
                        pass
                else:
                    best_binding_local[antigen] = significance_binding

            sorted_keys = sorted(best_binding_local, key=best_binding_local.get, reverse=True)

            # Get the best and second best keys
            best_key = sorted_keys[0]
            if len(found_antigens) >= 3:
                second_best_key = sorted_keys[1] if len(sorted_keys) > 1 else None
                best_binding[i] = [best_key, second_best_key]
            else:
                best_binding[i] = [best_key]
        return best_binding


    def full_analysis(self):
        """
        :return: returns some plots about quality in /my_experiments/YOUR_EXPERIMENT_NAME/plots. Is called automatically when launching the pipeline.
        """
        try:
            mh_save_path = os.path.join(self.report_path, "morosita_horn_identity" + ".xlsx")
            self.morosita_horn(matrix_save_path = mh_save_path)
            plt.tight_layout()
            self.save_in_plots("morosita_horn")
        except:
            print("Creating Morosita Horn matrix failed")
        try:
            self.lengthDistribution_multi()
            plt.tight_layout()
            self.save_in_plots("length_distribution_multi")
        except:
            print("Creating Length Distribution of samples failed")
        try:
            print("Begin with creating mixcr plots. You can find a description about them here:\n https://mixcr.com/mixcr/reference/mixcr-exportPlots/")
            self.export_mixcr_quality()
            self.mixcr_explain_diversity()
            self.mixcr_vdj_usage()
            self.mixcr_overlap()
        except:
            print(f".clns files could not be found. Most certainly, you have uploaded the sequencing report.")
        try:
            self.alignment_quality()
            plt.tight_layout()
            self.save_in_plots("alignment_quality")
        except:
            print("Alignment reports could not be found")
        for i in self.experiments_list:
            self.rarefraction_curves(samples = [i])
            plt.tight_layout()
            self.save_in_plots("rarefraction_curves_"+i)
       # self.relative_abundance_multi()
       # plt.tight_layout()
        #self.save_in_plots("relative_sequence_abundance_multi")
        try:
            self.logoPlot_multi()
            plt.tight_layout()
            self.save_in_plots("logo_plot_multi")
        except:
            print("Creating Logo plot of multiple samples failed")
        print("Cluster NGS sequences in dendrograms and network plots using levenshtein distance of 2 for network plots and ls-distance of 1 for dendrogram. Batch size is set to 1000.")
        if not os.path.isdir(os.path.join(self.plot_path, "sequence_cluster")):
            os.mkdir(os.path.join(self.plot_path, "sequence_cluster"))

            for single_experiment in self.experiments_list:
                try:
                    self.levenshtein_dendrogram(sample = single_experiment, max_cluster_dist = 1)
                    self.save_in_plots(os.path.join("sequence_cluster", single_experiment + "ls_cluster"))
                    self.basic_cluster(samples = [single_experiment],
                                       max_ld = 2,
                                       batch_size=1000,
                                       save_report_path = os.path.join(self.report_path, f"{single_experiment}" + "ls_cluster" + ".xlsx"))
                    self.save_in_plots(os.path.join("sequence_cluster", single_experiment + "ls_cluster"))
                except:
                    print(f"Sequence clustering failed at: {single_experiment}")
        threshold_identity = 0.2
        while True:
            overlapping_samples = read_matrix(threshold=threshold_identity, path_matrix=mh_save_path)
            mean_length = sum(len(v) for v in overlapping_samples.values()) / len(overlapping_samples)
            if mean_length > 2.5:
                break
            elif threshold_identity < 0.05:
                break
            else:
                threshold_identity -= 0.01
        for single_experiment in list(overlapping_samples.keys()):
            try:
                self.embedding_tsne(samples = overlapping_samples[single_experiment])
                self.save_in_plots(single_experiment + "embedding_tsne")
            except:
                print(f"Clustering with TSNE failed for {single_experiment}.")
        best_binder = self.get_best_binder()    
        if best_binder == None:
            print("No binding data was found. Thus, no binding plots can be created.")
        else:
            clustering_antigens_path = os.path.join(self.plot_path, "clustering_antigens")
            if not os.path.isdir(clustering_antigens_path):
                os.mkdir(clustering_antigens_path)
            experiment_keys = list(best_binder.keys())
            print("Starting clustering of binding data for antigens and sequences using PCA and t-SNE.\nPerplexity is set to 25.\n70 principal components are taken for t-SNE.\n2500 iterations are used for tsne.")
            for experiment in experiment_keys:
                self.tsne_cluster_AG(sample = experiment,
                                     antigen = best_binder[experiment],
                                     save_report_path = os.path.join(self.report_path, experiment + "_best_binder[experiment]" + ".xlsx"))
                self.save_in_plots(os.path.join( "clustering_antigens", experiment + "tsne_cluster_AG"))
            dendro_path = os.path.join(self.plot_path,"clustering_antigens", "dendro_binding")
            if not os.path.isdir(dendro_path):
                os.mkdir(dendro_path)
            print("Create dendrogram with binding data and levenshtein distance of 2 and batch size of 1000.")
            for experiment in experiment_keys:
                fig2 = self.dendro_bind(sample = experiment,
                                 antigens = best_binder[experiment],
                                 )
                self.fig.savefig(os.path.join(self.module_dir, "my_experiments", self.experiment, "plots", "clustering_antigens", "dendro_binding", experiment + f"_{best_binder[experiment]}.png"))
                fig2.savefig(os.path.join(self.module_dir, "my_experiments", self.experiment, "plots", "clustering_antigens", "dendro_binding", experiment + f"_{best_binder[experiment]}_values.png"))


        print(f"The pipeline has created some plots in: {self.plot_path}")
        

    def change_preferred_antigen(self, antigen=None):
        if self.binding_data is not None:
            columns_bind_data = self.binding_data.columns.to_list()
            if antigen == None:
                antigen = columns_bind_data[1]
            self.preferred_antigen = antigen
        else:
            self.preferred_antigen = None

    def change_preferred_sample(self, sample):
        assert sample in self.experiments_list, "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        self.preferred_sample = sample

    def change_filter(self, divisible_by=3, length_threshold=9, min_read_count=0):
        """
        :param divisible_by: filter all sequences which are not divisible by this value
        :param length_threshold: filter out all sequences which are shorter than this value
        :min_read_count: filter out all sequences which have less clones/reads than this value
        """
        self.Report.prepare_seq_report(self.region_string, divisible_by, length_threshold, min_read_count)
        self.sequencing_report = self.Report.sequencing_report

    def change_region(self):
        """
        :return: changes the region you want to analyse
        """
        intermediate = self.region_of_interest
        possible_regions = self.avail_regions + ["all"]
        region_string = input(
            f"Which region do you want to plot? ExpoSeq could find the following regions: {self.avail_regions}. If you want to merge your sequences and analyse the longest possible consecutive sequences your dataset offers type 'all'")
        if region_string in possible_regions:
            if not region_string == "all":
                region_string = region_string.replace("nSeq", "")
                self.Report.prepare_seq_report(region_string, divisible_by=3, length_threshold=9, min_read_count=0)
            else:
                self.Report.filter_longest_sequence()

            self.sequencing_report = self.Report.sequencing_report
            self.region_of_interest = "aaSeq" + region_string
        else:
            print(f"The region you want to plot is not valid. The options are: {self.avail_regions}")
            self.region_of_interest = intermediate

    def discard_samples(self, samples_to_discard):
        """

        :param samples_to_discard: the name of the samples in a list which you want to delete from your further analysis
        :return: updates the sequencing report where the values for the given samples do not exist anymore

        """
        assert type(samples_to_discard) == list, "You have to give a list with the samples you want to discard"
        self.sequencing_report = self.sequencing_report[~self.sequencing_report['Experiment'].isin(samples_to_discard)]

    def add_binding_data(self):
        """"
        :return: adds binding data to your analysis. You can add mutliple files with the filechooser or a given path to the file. For more information about the necessary file structure, check the supplementary information.
        """
        self.binding_data = collect_binding_data(binding_data=self.binding_data)

        self.change_preferred_antigen(self.binding_data.columns.to_list()[1])

        # self.binding_data.to_csv("binding_data.csv")

    def merge_bind_seq_report(self):
        """
        :return: merges the binding data with the sequencing report.
        """
        assert self.binding_data is not None, "You have not given binding data. You can add it with the add_binding_data function"

        merged_reports = pd.merge(self.sequencing_report, self.binding_data, on="aaSeqCDR3", how="outer")
        return merged_reports

    def print_antigens(self):
        """
        :return: prints the antigens (columns) of your binding data
        """
        try:
            print(self.binding_data.columns.to_list()[1:])
        except:
            print("You have not given binding data.")

    def print_samples(self):
        """
        :return: prints the names of your samples, so you can insert them in lists or similar for the analysis with some plots
        """
        print(list(self.unique_experiments.values()))

    def save(self, enter_filename, dpi=300, format="png"):
        """
        :return: can be used to save your plots. It will ask you automatically for the directory where you want to save it
        """
        print("Choose the directory where you want to save your plot with the filechooser")
        try:
            from tkinter import filedialog
            save_dir = filedialog.askdirectory()
        except:
            save_dir = ""

        plt.savefig(fname=os.path.join(save_dir, enter_filename + f".{format}"), dpi=dpi, format=format)

    def close(self):
        plt.close()

    def change_experiment_names(self, specific=None, change_whole_dic=False):
        """
        :param specific: optional parameter. You can use this function to change the names of a specific sample.
        :param change_whole_dic: optional Parameter. The renaming can be done by using a dictionary and map it to the labels. You can change multiple or all labels by adding the new dictionary for this parameter.
        :return:You can use this function to change the name of your samples. Thus, you can change the labels of your plots.
        """

        self.unique_experiments = self.sequencing_report["Experiment"].unique().tolist()
        self.unique_experiments = dict(zip(self.unique_experiments, self.unique_experiments))
        if specific != None:
            assert type(
                specific) == str, "You have to give a string (inside: "") as input for the specific sample you want to change"
        if change_whole_dic != False:
            assert type(
                change_whole_dic) == dict, "You have to give a dictionary as input for the specific sample you want to change"
        if change_whole_dic == False:
            if specific == None:
                print(
                    "You will no go through all your samples and you will be able to change their names. If you want to skip a samples, just press enter.")
                for key in self.unique_experiments:
                    print(f"Current value for {key}: {self.unique_experiments[key]}")
                    new_value = input("Enter a new value or press any key to skip")
                    if len(new_value) > 1:
                        self.unique_experiments[key] = new_value
                    else:
                        pass
            else:
                new_value = input("Enter the new name for " + specific)
                self.unique_experiments[specific] = new_value
            self.sequencing_report["Experiment"] = self.sequencing_report["Experiment"].map(self.unique_experiments)
        else:
            with open("test_data" + "/experiment_names.pickle", "wb") as f:
                pickle.dump(change_whole_dic, f)

    def alignment_quality(self, log_transformation=False):
        """

        :param log_transformation: optional parameter. You can set it to True if you want to log transform your data
        :return: Creates a figure where you can see the Overall Reads per sample and the number of reads which could be aligned.
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        assert log_transformation in [True, False], "You have to give True or False as input for the log transformation"
        self.ControlFigure.clear_fig()
        try:
            barplot.barplot(self.ControlFigure.ax,
                            self.alignment_report,
                            self.sequencing_report,
                            self.font_settings,
                            self.legend_settings,
                            apply_log=log_transformation)

            self.ControlFigure.update_plot()
            self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                               self.ControlFigure.plot_type)
        except:
            print("Currenlty you do not have loaded an alignment report.")

    def aa_distribution(self, sample=None, region=[3, 7], protein=True):
        """
        :param sample: The sample you would like to analyze
        :param region: The input is a list and you can specify with the first value the beginning region and with the second value the end region within your sequences for which you want to know the amino acid composition. For instance: [3,7]
        :param protein: Default True. If you would like to analyze nucleotide sequences, set it to False
        :return: Returns a plot which shows the amino acid distribution in the given sequence range.
        """
        if sample == None:
            sample = self.preferred_sample
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        assert sample in self.experiments_list, "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        assert type(sample), "You have to give a string as input for the sample"
        assert type(
            region), "You have to give a list with the start and end position of the region you want to analyze. For instance: [3,7]"
        assert protein in [True, False], "You have to give True or False as input for the protein parameter"
        self.ControlFigure.clear_fig()
        stacked_aa_distribution.stacked_aa_distr(self.ControlFigure.ax,
                                                 self.sequencing_report,
                                                 sample,
                                                 region,
                                                 protein,
                                                 self.font_settings,
                                                 self.region_of_interest)

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)

    def rarefraction_curves(self, samples=None):
        """
        :param samples: you insert a list which contains the sample names
        :return: Shows your the rarefraction curves for the given samples.
        """
        if samples == None:
            samples = [self.experiments_list[0]]
        incorrect_samples = [x for x in samples if x not in self.experiments_list]
        assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"

        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if type(samples) != list:
            print(
                "You have to give the sample names in the list, also if it is only one! A list is a container which is marked through:[] . Please try again.")
        else:
            self.ControlFigure.clear_fig()
            usq_plot.plot_USQ(fig=self.ControlFigure.fig,
                              sequencing_report=self.sequencing_report,
                              samples=samples,
                              font_settings=self.font_settings,
                              legend_settings=self.legend_settings)

            self.ControlFigure.update_plot()
            self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                               self.ControlFigure.plot_type)

    def logoPlot_single(self,
                        sample=None,
                        highlight_specific_pos=False,
                        method="proportion",
                        chosen_seq_length=16):
        """
        :param sample: insert the sample name
        :param highlight_specific_pos: optional. you can highlight a specific position. For instance if you want to highlight the 3rd position, you insert 3.
        :param method: You can specify whether you want to have on your y axis the frequency of the amino acids or the information content in bits. The default is proportion. If you want to have the information content, insert "bits"
        :param chosen_seq_length: 16 per default. You always analyze online one sequence length! You can change it if you would like to.
        :return: A logo Plot which shows you the composition of aminoacids per position
        """
        if sample == None:
            sample = self.preferred_sample
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        assert sample in self.experiments_list, "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        assert type(sample) == str, "You have to give a string as input for the sample"
        if highlight_specific_pos != False:
            assert type(
                highlight_specific_pos) == int, "You have to give an integer as input for the specific position you want to highlight"
        assert type(
            chosen_seq_length) == int, "You have to give an integer as input for the sequence length you want to analyze"
        self.ControlFigure.clear_fig()
        logo_plot.plot_logo_single(self.ControlFigure.ax,
                                   self.sequencing_report,
                                   sample,
                                   self.font_settings,
                                   highlight_specific_pos,
                                   self.region_of_interest,
                                   method,
                                   chosen_seq_length)

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)

    def logoPlot_multi(self,
                       samples="all",
                       chosen_seq_length=16,
                       method="proportion",
                       ):
        """
        :param samples: You analyze all samples per default. If you want to analyze specific samples it has to be a list with the corresponding sample names
        :param chosen_seq_length: 16 per default. You always analyze online one sequence length! You can change it if you would like
        :return: Gives you in one figure one logoPlot per sample.
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        if samples != "all":
            assert type(samples) == list, "You have to give a list with the samples you want to analyze"
            incorrect_samples = [x for x in samples if x not in self.experiments_list]
            assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        assert type(
            chosen_seq_length) == int, "You have to give an integer as input for the sequence length you want to analyze"
        self.ControlFigure.clear_fig()
        logo_plot.plot_logo_multi(self.ControlFigure.fig,
                                  self.sequencing_report,
                                  samples,
                                  self.font_settings,
                                  self.region_of_interest,
                                  method,
                                  chosen_seq_length,
                                  )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)
        self.ControlFigure.tighten()

    def lengthDistribution_single(self, sample=None):
        """

        :param sample: name of the sample from which you would like to see the length distribution
        :return: Shows you the length Distribution of your sequences of the given sample
        """
        if sample == None:
            sample = self.preferred_sample
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        assert type(sample) == str, "You have to give a string as input for the sample"
        assert sample in self.experiments_list, "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        length_distribution.length_distribution_single(self.ControlFigure.fig,
                                                       self.ControlFigure.ax,
                                                       self.sequencing_report,
                                                       sample,
                                                       self.font_settings,
                                                       self.region_of_interest)

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)

    def relative_abundance_multi(self, samples="all"):
        """
        :param samples: You analyze all samples per default. If you want to analyze specific samples it has to be a list with the corresponding sample names
        :return: Outputs a figure which shows the fractions per clones for each sample in the dataset
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"

        if samples != "all":
            assert type(samples) == list, "You have to give a list with the samples you want to analyze"
            incorrect_samples = [x for x in samples if x not in self.experiments_list]
            assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        else:
            samples = self.experiments_list
        self.ControlFigure.clear_fig()
        relative_sequence_abundance.relative_sequence_abundance_all(self.ControlFigure.fig,
                                                                    self.sequencing_report,
                                                                    samples,
                                                                    self.font_settings,
                                                                    self.region_of_interest)
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)
        self.ControlFigure.tighten()

    def lengthDistribution_multi(self, samples="all"):
        """
        :param samples: You analyze all samples per default. If you want to analyze specific samples it has to be a list with the corresponding sample names
        :return: Outputs one figure with one subplot per sample which shows you the distribution of the sequence length
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        if samples != "all":
            assert type(samples) == list, "You have to give a list with the samples you want to analyze"
            incorrect_samples = [x for x in samples if x not in self.experiments_list]
            assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        length_distribution.length_distribution_multi(self.ControlFigure.fig,
                                                      self.sequencing_report,
                                                      samples,
                                                      self.font_settings,
                                                      self.region_of_interest,
                                                      test_version=self.is_test)
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)
        self.ControlFigure.tighten()

    def rel_seq_abundance(self, samples=None, max_levenshtein_distance=0, length_filter=0, batch=3000):
        """

        :param samples: For a qualitative analysis choose samples from the same panning experiment. Input is a list
        :param max_levenshtein_distance: Default is 0. You can change it to see increased fraction with increased variability of certain sequences
        :param length_filter: Default is 0. You should change it if you change the levenshtein distance. Otherwise your results will be biased.
        :param batch: Default is 3000. The size of the sample which is chosen. The higher it is, the more computational intense.
        :return: Shows you a bar plot of the frequencies of the most abundant sequences. You can introduce the levenshtein distance to see how the frequency changes with higher variability of the sequences.
        """
        if samples == None:
            samples = [self.experiments_list[0]]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        incorrect_samples = [x for x in samples if x not in self.experiments_list]
        assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        assert type(samples) == list, "You have to give a list with the samples you want to analyze"
        assert type(
            max_levenshtein_distance), "You have to give an integer as input for the maximum levenshtein distance"
        assert type(length_filter), "You have to give an integer as input for the length filter"
        assert type(batch) == int, "You have to give an integer as input for the batch size"
        self.ControlFigure.clear_fig()
        relative_sequence_abundance.relative_sequence_abundance(self.ControlFigure.ax,
                                                                self.sequencing_report,
                                                                samples,
                                                                max_levenshtein_distance,
                                                                length_filter,
                                                                batch,
                                                                self.region_of_interest,
                                                                self.font_settings,
                                                                self.legend_settings)

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)

    def save_cluster_report(self,cluster_report, path=None):
        if path == None:
            while True:
                cluster_report_save = input("Do you want to save the generated data? (Y/n)")
                if cluster_report_save in ["Y", "y", "N", "n"]:
                    break
                else:
                    print("Please enter Y or n")
            if cluster_report_save.lower() in ["Y", "y"]:
                while True:
                    filename_matrix = input("Enter a name for the file. The file will be saved locally in your IDE.")
                    if not os.path.isfile(filename_matrix):
                        path = os.path.join(self.report_path, filename_matrix)
                        cluster_report.to_excel(path)
                        break
                    else:
                        print("This file already exists. Please choose another name.")
        else:
            cluster_report.to_excel(path)


    def basic_cluster(self, samples = None, batch_size = 1000, max_ld = 1, min_ld = 0, second_figure = False, label_type = "numbers", save_report_path = None):
        """
        :param samples: A list containing the samples you would like to analyze. Analyze just one sample with: ["My_sampel_name"].
        :param max_ld: Maximum allowed levenshtein distance between sequences within one cluster. The higher the distance the larger the clusters.
        :param min_ld: Minimum allowed levenshtein distance between sequences within one cluster.
        :param second_figure: Default is False. Creates a second figure with the levenshtein distance as bars
        :param label_type: Default is numbers. This will label the nodes in the plot with the corresponding identifier in the output report. You can type sequences for labeling the nodes with the sequneces. If you do not want to have labels set it to None.
        :param save_report_path: Default is None which saves your report in my_experiments/reports_pipeline. If you want to change it somewhere else you need to insert the full path with filename.
        """
        if samples == None:
            samples = [self.preferred_sample]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
      #  assert sample in self.experiments_list, "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        assert type(samples) == list, "You have to give a string as input for the sample"
        assert type(max_ld) == int, "You have to give an integer as input for the maximum levenshtein distance"
        assert type(min_ld) == int, "You have to give an integer as input for the minimum levenshtein distance"
        assert type(second_figure) == bool, "You have to give True or False as input for the second figure"
        self.ControlFigure.clear_fig()
        cluster_report = levenshtein_clustering.clusterSeq(
            self.ControlFigure.ax,
            self.sequencing_report,
            samples,
            max_ld,
            min_ld,
            batch_size,
            self.font_settings,
            second_figure,
            self.region_of_interest,
            label_type)

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)
        self.save_cluster_report(cluster_report = cluster_report,
                                 path = save_report_path)
        if second_figure == True:
            print("close the second window before you continue")


    def cluster_one_AG(self, antigen=None, max_ld=1, min_ld=0, batch_size=1000, specific_experiments=False, preferred_cmap = "Blues", label_type = "numbers", save_report_path = None):
        """
        :param antigen: is the name of the antigen you would like to analyze
        :param max_ld: optional Parameter where its default is 1. Is the maximum Levenshtein distance you allow per cluster
        :param min_ld: optional Parameter where its default is 0. Is the minimum Levenshtein distance between sequences you allow
        :param batch_size: optional Parameter where its default is 1000. Is the batch size you want to use for the analysis
        :param specific_experiments: optional Parameter. You can give the names of specific samples in a list if you want
        :param preferred_cmap: optional Parameter. Define the colormap you would like to use.
        :param save_report_path: Default is None which saves your report in my_experiments/reports_pipeline. If you want to change it somewhere else you need to insert the full path with filename.
        :return: Creates a figure where sequences are clustered based on Levenshtein distance. Additionally the binding data of the sequences against a specific antigen is given.
        """

        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        assert self.binding_data is not None, "You have not given binding data. You can add it with the add_binding_data function"
        if antigen == None:
            antigen = self.preferred_antigen
        assert type(antigen) == str, "You have to give a string as input for the antigen"
        assert type(max_ld) == int, "You have to give an integer as input for the maximum levenshtein distance"
        assert type(min_ld) == int, "You have to give an integer as input for the minimum levenshtein distance"
        assert type(specific_experiments) == list, "You have to give a list with the samples you want to analyze"
        self.ControlFigure.clear_fig()
        cluster_report = levenshtein_clustering.cluster_single_AG(self.ControlFigure.fig,
                                                 self.sequencing_report,
                                                 antigen,
                                                 self.binding_data,
                                                 max_ld,
                                                 min_ld,
                                                 batch_size,
                                                 self.region_of_interest,
                                                 preferred_cmap,
                                                 specific_experiments,
                                                 label_type
                                                 )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)
        self.save_cluster_report(cluster_report = cluster_report,
                                 path = save_report_path)

    def tsne_cluster_AG(self, sample=None, antigen=None, antigen_names=True, pca_components=70, perplexity=25,
                        iterations_tsne=2500, save_report_path = None):
        """
        :param sample: the sample you would like to analyze
        :param antigen: the toxins you would like to cluster
        :param antigen_names: Default is True. Prints the name of the toxin for the corresponding embedded sequence in the plot
        :param pca_components: optional. Default is 70
        :param perplexity: optional. Default 25
        :param iterations_tsne: optional. Default is 2500
        :return: It first embeds the sequences in a vector space and then reduces the dimensions and clusters them with PCA and TSNE. The sequences with the binding data are processed with the input sequences, to enable the plotting of the binding data.
        """
        if sample == None:
            sample = self.preferred_sample
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        assert self.binding_data is not None, "You have not given binding data. You can add it with the add_binding_data function"
        if antigen == None:
            antigen = [self.preferred_antigen]
        assert type(sample) == str, "You have to give a string as input for the sample"
        assert sample in self.experiments_list, "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        assert antigen_names in [True, False], "You have to give True or False as input for the antigen_names parameter"
        assert type(pca_components) == int, "You have to give an integer as input for the pca_components"
        assert type(perplexity) == int, "You have to give an integer as input for the perplexity"
        assert type(iterations_tsne) == int, "You have to give an integer as input for the iterations_tsne"
        assert type(antigen) == list, "You have to give a list with the antigens you want to analyze"

        self.ControlFigure.clear_fig()
        tsne_results = embedding_with_binding.cluster_toxins_tsne(self.ControlFigure.fig,
                                                                  self.sequencing_report,
                                                                  sample,
                                                                  antigen,
                                                                  self.binding_data,
                                                                  antigen_names,
                                                                  pca_components,
                                                                  perplexity,
                                                                  iterations_tsne,
                                                                  self.font_settings,
                                                                  self.colorbar_settings,
                                                                  False,
                                                                  self.region_of_interest)
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)

        self.save_cluster_report(tsne_results, path = save_report_path)

    def embedding_tsne(self,
                       samples=None,
                       strands=True,
                       pca_components=80,
                       perplexity=30,
                       iterations_tsne=2500,
                       batch_size=1000):
        """
        :param samples: the samples you would like to compare towards their sequences
        :param strands: Default is True. It means that you will plot a batch of the strands in your plot
        :param pca_components: Default is 80. Has to be applied for better accuracy of t-SNE. You can indirectly change the described variance with this.
        :param perplexity: Default is 30. It roughly determines the number of nearest neighbors that are considered in the embedding. A higher perplexity value results in a more global structure in the low-dimensional embedding, while a lower perplexity value emphasizes local structure. The optimal perplexity value for a given dataset depends on the dataset's intrinsic dimensionality, and it is usually determined by trial and err
        :param iterations_tsne: Default is 2500. number of times that the algorithm will repeat the optimization process for reducing the cost function. The optimization process aims to minimize the difference between the high-dimensional and low-dimensional representations of the data. More iterations result in a more optimized low-dimensional representation, but also increases the computational cost.
        :param batch_size: Default is 1000. The size of the sample which is chosen. The higher it is, the more computational intense.
        :return: Returns a plot where the sequences of the input samples are transformed in a vector space. Dimension reduction such as PCA and following t-SNE is used to plot it on a two dimensional space. The different colors indicate the different samples.
        """
        if samples == None:
            samples = [self.experiments_list[0]]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        incorrect_samples = [x for x in samples if x not in self.experiments_list]
        assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        assert type(samples) == list, "You have to give a list with the samples you want to analyze"
        assert type(strands) == bool, "You have to give True or False as input for the strands parameter"
        assert type(pca_components) == int, "You have to give an integer as input for the pca_components"
        assert type(perplexity) == int, "You have to give an integer as input for the perplexity"
        assert type(iterations_tsne) == int, "You have to give an integer as input for the iterations_tsne"
        self.ControlFigure.clear_fig()
        cluster_embedding.show_difference(self.sequencing_report,
                                          samples,
                                          strands,
                                          batch_size,
                                          pca_components,
                                          perplexity,
                                          iterations_tsne,
                                          self.region_of_interest,
                                          self.ControlFigure.ax,
                                          self.legend_settings,
                                          self.font_settings)
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)

    def morosita_horn(self, annotate_cells=False, specific_experiments=False, matrix_save_path = None):
        """
        :param annotate_cells: Default is False. If you want to see the values of the matrix, set it to True.
        :param specific_experiments: you can give a list with specific experiments
        :param matrix_save_path: Path where you want to save the matrix
        :return: Returns a matrix of the identity between your samples based on the Morosita Horn Index
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if specific_experiments != False:
            assert type(specific_experiments) == list, "You have to give a list with the samples you want to analyze"
            incorrect_samples = [x for x in specific_experiments if x not in self.experiments_list]
            assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        matrix = plt_heatmap.plot_heatmap(self.sequencing_report,
                                          True,
                                          "morosita_horn",
                                          self.ControlFigure.ax,
                                          self.colorbar_settings,
                                          self.font_settings,
                                          annotate_cells,
                                          self.region_of_interest,
                                          specific_experiments=specific_experiments,
                                          )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)
        save_matrix(matrix, matrix_save_path)

    def jaccard(self, annotate_cells=False, specific_experiments=False):
        """
        :param annotate_cells: Default is False. If you want to see the values of the matrix, set it to True.
        :param specific_experiments: give a list with specific samples you would like to analyze
        :return: Returns a matrix of the identity between your samples based on the Jaccard Index
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if specific_experiments != False:
            assert type(specific_experiments) == list, "You have to give a list with the samples you want to analyze"
            incorrect_samples = [x for x in specific_experiments if x not in self.experiments_list]
            assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        matrix = plt_heatmap.plot_heatmap(self.sequencing_report,
                                          True,
                                          "jaccard",
                                          self.ControlFigure.ax,
                                          self.colorbar_settings,
                                          self.font_settings,
                                          annotate_cells,
                                          self.region_of_interest,
                                          specific_experiments=specific_experiments,
                                          )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)
        save_matrix(matrix)

    def sorensen(self, annotate_cells=False, specific_experiments=False):
        """
        :param annotate_cells: Default is False. If you want to see the values of the matrix, set it to True.
        :param specific_samples: give a list with specific samples you would like to analyze
        :return: Returns a matrix of the identity between your samples based on the Sorensen Dice Index
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if specific_experiments != False:
            assert type(specific_experiments) == list, "You have to give a list with the samples you want to analyze"
            incorrect_samples = [x for x in specific_experiments if x not in self.experiments_list]
            assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        matrix = plt_heatmap.plot_heatmap(self.sequencing_report,
                                          True,
                                          "sorensen",
                                          self.ControlFigure.ax,
                                          self.colorbar_settings,
                                          self.font_settings,
                                          annotate_cells,
                                          self.region_of_interest,
                                          specific_experiments=specific_experiments,
                                          )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)
        save_matrix(matrix)

    def relative(self, annotate_cells=False, specific_experiments=False):
        """
        :param annotate_cells: Default is False. If you want to see the values of the matrix, set it to True.
        :param specific_samples: give a list with specific samples you would like to analyze
        :return: Returns a matrix of the identity between your samples based on the proportion of identical sequences
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if specific_experiments != False:
            assert type(specific_experiments) == list, "You have to give a list with the samples you want to analyze"
            incorrect_samples = [x for x in specific_experiments if x not in self.experiments_list]
            assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        matrix = plt_heatmap.plot_heatmap(self.sequencing_report,
                                          True,
                                          "relative",
                                          self.ControlFigure.ax,
                                          self.colorbar_settings,
                                          self.font_settings,
                                          annotate_cells,
                                          self.region_of_interest,
                                          specific_experiments=specific_experiments,
                                          )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)
        save_matrix(matrix)

    def levenshtein_dendrogram(self, sample=None, max_cluster_dist=2, batch_size=1000):
        """
        :params sample: the sample you would like to analyze
        :max_cluster_dist: Default is 2. Maximum levenshtein distance between sequences within a cluster.
        :params batch_size: Default is 1000. The size of the sample which is chosen.
        :return: Returns a dendrogram of the sequences based on the Levenshtein distance
        """
        if sample == None:
            sample = self.preferred_sample
        assert type(sample) == str, "You have to give a string as input for the sample"
        assert sample in self.experiments_list, "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        assert type(
            max_cluster_dist) == int, "You have to give an integer as input for the maximum levenshtein distance"
        assert type(batch_size) == int, "You have to give an integer as input for the batch size"
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        self.ControlFigure.clear_fig()
        hist_lvst_dist.levenshtein_dend(self.ControlFigure.ax,
                                        self.sequencing_report,
                                        sample,
                                        batch_size,
                                        max_cluster_dist,
                                        self.font_settings,
                                        self.region_of_interest
                                        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)

    def dendro_bind(self, sample=None, antigens=None, max_cluster_dist=2, batch_size=1000, ascending=True):
        """
        :params sample: the sample you would like to analyze
        :params antigens: the antigens you would like to analyze. The input is a list.
        :max_cluster_dist: Default is 2. Maximum levenshtein distance between sequences within a cluster. The higher this number the bigger the dendrogram
        :params batch_size: Default is 1000. The number of sequences starting with the highest fractions which are used for the analysis from the sample
        :params ascending: Default is True. If you want to see the highest fractions first, set it to False
        :return: Creates a dendrogram which shows the realtionship between your sequences in the sample and your sequences with binding data based on levenshtein distance. Additionally a barplot with the binding data of the found sequences is given.
        """
        if sample == None:
            sample = self.preferred_sample
        assert self.binding_data is not None, "You have not given binding data. You can add it with the add_binding_data function"
        if antigens == None:
            antigens = [self.preferred_antigen]
        assert type(sample) == str, "You have to give a string as input for the sample"
        assert sample in self.experiments_list, "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        assert type(
            max_cluster_dist) == int, "You have to give an integer as input for the maximum levenshtein distance"
        assert type(batch_size) == int, "You have to give an integer as input for the batch size"
        assert type(ascending) == bool, "You have to give True or False as input for the ascending parameter"

        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        self.ControlFigure.clear_fig()
        fig2 = hist_lvst_dist.dendo_binding(self.ControlFigure.fig,
                                     self.sequencing_report,
                                     self.binding_data,
                                     sample,
                                     antigens,
                                     batch_size,
                                     max_cluster_dist,
                                     self.font_settings,
                                     self.region_of_interest,
                                     ascending
                                     )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(self.ControlFigure.ax,
                                           self.ControlFigure.plot_type)
        self.ControlFigure.tighten()
        return fig2

    def MSA_design(self, samples=None, batch_size=500):
        """
        :params samples: the samples you would like to analyze. The input is a list.
        :params batch_size: Default is 1000. The number of sequences starting with the highest fractions which are used for the analysis from the sample
        :return: Creates a multiple sequence alignment of the sequences of the samples you have chosen. The sequences are ordered by their frequency and they are chosen equally based on the given batch size from the different samples. The output is an interface with an overview of the MSA and the option to design a sequence based on the MSA and the findings of the analysis.
        """
        if samples == None:
            samples = [self.experiments_list[0]]
        assert type(samples) == list, "You have to give a list with the samples you want to analyze"
        incorrect_samples = [x for x in samples if x not in self.experiments_list]
        assert not incorrect_samples, f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"

        msa = MSA(self.region_of_interest,
                  self.Report,
                  self.Settings,
                  self.module_dir)
        msa = msa.run_MSA(batch_size, samples)

    def validate_mixcr_path(self):
        path_to_mixcr = self.global_params["mixcr_path"]
        settings_dir = os.path.join(self.module_dir, "settings", "global_vars.txt")
        check_mixcr(path_to_mixcr, self.global_params, settings_dir)

    def create_parser(self):
        self.validate_mixcr_path()
        path_to_mixcr = self.global_params["mixcr_path"]
        commands = []
        commands.extend(["java", f"-Xms{500}M", "-jar"])  # enable change of para
        commands.extend([path_to_mixcr])
        return commands

    def export_mixcr_quality(self,save_dir = None, specific_chain=None, metric=None, plot_type=None):
        if save_dir == None:
            save_dir = self.mixcr_plots_path
        output_file = os.path.join(save_dir, self.experiment + "_alignQc.pdf")
        path_to_tables = os.path.join(self.module_dir, "my_experiments", self.experiment, "clones_result", "*.clns")
        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportQc"])
        export_plots_commands.extend(["align"])
        export_plots_commands.extend([path_to_tables])
        export_plots_commands.extend([output_file])
        export_plots_commands.extend(["--force-overwrite"])
        subprocess.run(export_plots_commands)
        print(f"You can find the result file at: {output_file}")

        for i in glob.glob(path_to_tables):

            name=os.path.basename(i).split(".")[0]
           # output_file = os.path.join(save_dir, name + "_tags.pdf")
            #export_plots_commands = self.create_parser()
            #export_plots_commands.extend(["exportQc"])
            #export_plots_commands.extend(["tags"])
            #export_plots_commands.extend([i])
            #export_plots_commands.extend([output_file])
            #export_plots_commands.extend(["--force-overwrite"])
            #subprocess.run(export_plots_commands)
            #print(f"You can find the result file at: {output_file}")
            name = os.path.basename(i).split(".")[0]
            output_file = os.path.join(save_dir, name + "chainUsage.pdf")
            export_plots_commands = self.create_parser()
            export_plots_commands.extend(["exportQc"])
            export_plots_commands.extend(["chainUsage"])
            export_plots_commands.extend(["--hide-non-functional"])
            export_plots_commands.extend([i])
            export_plots_commands.extend([output_file])
            export_plots_commands.extend(["--force-overwrite"])
            subprocess.run(export_plots_commands)
            print(f"You can find the result file at: {output_file}")
        output_file = os.path.join(save_dir, self.experiment + "_coverage.pdf")
        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportQc"])
        export_plots_commands.extend(["coverage"])
        export_plots_commands.extend([path_to_tables])
        export_plots_commands.extend([output_file])
        export_plots_commands.extend(["--force-overwrite"])
        subprocess.run(export_plots_commands)
        print(f"You can find the result file at: {output_file}")

        if specific_chain != None:
            export_plots_commands.extend(["--chains", specific_chain])
        if metric != None:
            export_plots_commands.extend(["--metrics", metric])
        if plot_type != None:
            export_plots_commands.extend(["--plotType", plot_type])


    def mixcr_explain_diversity(self,plot_name = "cdr3_metrics", plot_save_dir = None, metric = "cdr3metrics",chains = None, plot_type = "violin", output_type = "pdf"):
        """
        :param: metric: Default is cdr3metrics. You can choose between cdr3metrics and diversity. If you choose diversity you will focus on the clone fractions whereas with cdr3metrics you will focus on certain aspects of these chains such as length (https://mixcr.com/mixcr/reference/mixcr-postanalysis/#diversity-measures)
        :param: chains: Default is None. Possible inputs are : IGH, IGL, IGK, TRA, TRB, TRG, TRD, IG. You will choose a specific chain for the analysis.
        :param: plot_type: Default is violin. You can choose between: boxplot, boxplot-bindot, boxplot-jitter, violin, violin-bindot, barplot, barplot-stacked, lineplot, lineplot-jitter, lineplot-bindot, scatter
        :param: output_type: Default is pdf. Allows you to control the file type of your output. Possible values are: pdf, jpeg, png, svg, eps,
        """
        if metric in ["cdr3metrics", "diversity"]:
            pass
        else:
            print("Please choose a correct value for the metric. Pipeline continues with:\ncdr3metrics and cdr3lenAA for metric_type")
            metric = "cdr3metrics"
            metric_type = "cdr3lenAA"

        chains_values = [None, "IGH", "IGL", "IGK", "TRA", "TRB", "TRG", "TRD", "IG"]
        if chains not in chains_values:
            print("Please enter a correct shortcut for the chain preferred chain. These are: ")
            [print(i) for i in chains_values]
        plot_type_values = ["boxplot", "boxplot-bindot", "boxplot-jitter", "violin", "violin-bindot", "barplot", "barplot-stacked", "lineplot", "lineplot-jitter", "lineplot-bindot", "scatter"]
        if plot_type not in plot_type_values:
            print("Please enter a correct shortcut for the chain plot type. These are: ")
            [print(i) for i in plot_type_values]


        print(f"To be able to understand the following processing go to: https://mixcr.com/mixcr/reference/mixcr-postanalysis/#overlap-postanalysis")
        plot_name = plot_name

        if plot_save_dir == None:
            plot_save_dir = self.mixcr_plots_path
        save_dir = os.path.join(self.module_dir, "my_experiments", self.experiment)
        json_dir = os.path.join(save_dir, "pa_individual.json")
        path_to_tables = os.path.join(self.module_dir, "my_experiments", self.experiment, "clones_result", "*.clns")
        if not os.path.isfile(json_dir):
            export_plots_commands = self.create_parser()
            export_plots_commands.extend(["postanalysis"])
            export_plots_commands.extend(["individual"])
            export_plots_commands.extend(["--default-downsampling", "none"])
            export_plots_commands.extend(["--default-weight-function", "none"])
            export_plots_commands.extend([path_to_tables])
            export_plots_commands.extend([json_dir])
            subprocess.run(export_plots_commands)
        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportPlots"])
        export_plots_commands.extend([metric])
      #  export_plots_commands.extend(["--metric",metric_type])
      #  export_plots_commands.extend(["-f"])
        if not chains == None:
            export_plots_commands.extend(["--chains", chains])
        export_plots_commands.extend(["--plot-type", plot_type])
        export_plots_commands.extend([json_dir])
        export_plots_commands.extend([os.path.join(plot_save_dir, plot_name + "." + output_type)])
        subprocess.run(export_plots_commands)



    def mixcr_vdj_usage(self,):
        save_dir = self.mixcr_plots_path

        json_dir = os.path.join(self.module_dir, "my_experiments", self.experiment, "pa_individual.json")
        if not os.path.isfile(json_dir):
            export_plots_commands = self.create_parser()
            export_plots_commands.extend(["postanalysis"])
            export_plots_commands.extend(["individual"])
            export_plots_commands.extend(["--default-downsampling", "none"])
            export_plots_commands.extend(["--default-weight-function", "none"])
            export_plots_commands.extend([path_to_tables])
            export_plots_commands.extend([json_dir])
            subprocess.run(export_plots_commands)

        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportPlots"])
        export_plots_commands.extend(["vUsage"])
      #  export_plots_commands.extend(["-f"])
        export_plots_commands.extend([json_dir])
        export_plots_commands.extend([os.path.join(save_dir, "vUsage_heatmap.pdf")])
        subprocess.run(export_plots_commands)
        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportPlots"])
        export_plots_commands.extend(["jUsage"])
       # export_plots_commands.extend(["-f"])
        export_plots_commands.extend([json_dir])
        export_plots_commands.extend([os.path.join(save_dir, "jUsage_heatmap.pdf")])
        subprocess.run(export_plots_commands)
        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportPlots"])
        export_plots_commands.extend(["vjUsage"])
       # export_plots_commands.extend(["-f"])
        export_plots_commands.extend([json_dir])
        export_plots_commands.extend([os.path.join(save_dir, "vjUsage_heatmap.pdf")])
        subprocess.run(export_plots_commands)
        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportPlots"])
        export_plots_commands.extend(["isotypeUsage"])
       # export_plots_commands.extend(["-f"])
        export_plots_commands.extend([json_dir])
        export_plots_commands.extend([os.path.join(save_dir, "isotypeUsage_heatmap.pdf")])
        subprocess.run(export_plots_commands)


    def mixcr_overlap(self):
        save_dir = self.plot_path
        path_to_tables = os.path.join(self.module_dir, "my_experiments", self.experiment, "clones_result")
        json_dir = os.path.join(self.module_dir, "my_experiments", self.experiment, "pa_overlap.json")
        if not os.path.isfile(json_dir):
            export_plots_commands = self.create_parser()
            export_plots_commands.extend(["postanalysis"])
            export_plots_commands.extend(["overlap"])
            export_plots_commands.extend(["--default-downsampling", "none"])
            export_plots_commands.extend(["--default-weight-function", "none"])
            export_plots_commands.extend([path_to_tables])
            export_plots_commands.extend([json_dir])
            subprocess.run(export_plots_commands)
        possible_metrics = ["SharedClonotypes", "RelativeDiversity", "F1Index", "F2Index", "JaccardIndex", "Pearson", "PearsonAll"]

        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportPlots"])
        export_plots_commands.extend(["overlap"])
   #     export_plots_commands.extend(["--metric", metric])
       # export_plots_commands.extend(["--color-key", "Patient"])
        export_plots_commands.extend([json_dir])
        export_plots_commands.extend([os.path.join(save_dir, f"overlap.pdf")])
        subprocess.run(export_plots_commands)









import numpy as np
import editdistance


def create_distance_matrix(aa, max_dist=20):  # max_dist is a large finite value
    num_sequences = len(aa)
    distance_matrix = np.full((num_sequences, num_sequences), max_dist)  # initialize with max_dist

    for i in range(num_sequences):
        for j in range(i+1, num_sequences):
            lev_distance = editdistance.distance(aa[i], aa[j])
            if lev_distance < 2:
                distance_matrix[i, j] = lev_distance
                distance_matrix[j, i] = lev_distance  # The distance matrix is symmetric

    return distance_matrix

def get_clustered_sequences(aa):
    clustered_sequences = set()

    for i in range(len(aa)):
        for j in range(i+1, len(aa)):
            if editdistance.distance(aa[i], aa[j]) < 2:
                clustered_sequences.add(aa[i])
                clustered_sequences.add(aa[j])

    return list(clustered_sequences)



import subprocess
import os
from anno_sequence import SequenceManipulator

class Docker:
    def __init__(self):
        self.installed = False
        self.runs = False
    def is_docker_installed(self):
        try:
            subprocess.run(["docker", "--version"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            self.installed = True

        except subprocess.CalledProcessError:
            self.installed = False

    def is_docker_daemon_running(self):
        try:
            subprocess.run(["docker", "info"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            self.runs = True
        except subprocess.CalledProcessError:
            self.runs = False
            
    def install_message(self):
        if self.installed == False:
            print("Docker is not installed. Please install docker from: " + "https://docs.docker.com/get-docker/")
    
    def run_message(self):
        if self.runs == False:
            print("Docker is not running. Please check if you have created an account and that you are logged in.")

    def is_image_built(self, image_name ="alphafold"):
        cmd = ["docker", "images", "-q", image_name]
        result = subprocess.run(cmd, stdout=subprocess.PIPE, text=True)
        return bool(result.stdout.strip())



def run_alphafold(fasta_path, output_dir):
    cmd = [
        'docker', 'run',
        '-v', f"{fasta_path}:/data",
        '-v', f"{output_dir}:/output",
        'alphafold',
        '--fasta_paths=/data/my_sequence.fasta',
        '--max_template_date=YYYY-MM-DD',
        '--output_dir=/output'
    ]
    subprocess.run(cmd)



class DesignFasta():
    def __init__(self, template_sequence):
        self.reference = template_sequence
        self.new_query = template_sequence
        self.sequence_name = "my_sequence"
    
    def define_sequence_name(self):
        self.sequence_name = input("Please enter a name for your sequence: ")

    def manipulate_seq(self):
        app = SequenceManipulator(self.reference)
        app.run()
        self.new_query = app.sequence
        
    def write_fasta(self, module_dir):
        if not os.path.isdir("myfastasequences"):
            os.mkdir("myfastasequences")
        self.define_sequence_name()
        full_path = os.path.abspath(os.path.join(module_dir,"temp", self.sequence_name + ".fasta"))
        print(full_path)
        with open(full_path, "w") as f:
            f.write(">" + self.sequence_name + "\n")
            f.write(self.new_query.get())


def folding(template_seq, output_dir):
    module_dir = r"C:\Users\nilsh\my_projects\ExpoSeq"
    My_fasta = DesignFasta(template_seq)
    My_fasta.manipulate_seq()
    My_fasta.write_fasta(module_dir)
    filename = My_fasta.sequence_name
    My_docker = Docker()
    os.chdir(os.path.join(module_dir, "alphafold"))
    if My_docker.is_image_built():
        print(filename)
      #  print()
        subprocess.run(["python3",
                        r"C:\Users\nilsh\my_projects\ExpoSeq\alphafold\docker\run_docker.py",
                        "--fasta_paths=" + os.path.join(module_dir, "temp", filename + ".fasta"),
                        "--model_preset=monomer",
                        "--db_preset=reduced_dbs",
                        "--output_dir="+output_dir,
                        ])
     #   run_alphafold(os.path.join(module_dir, "temp", filename + ".fasta"), output_dir)
        print("Prediction_succesful")
    else:
        os.chdir(module_dir)
        My_docker.is_docker_installed()
        My_docker.install_message()
        My_docker.is_docker_daemon_running()
        My_docker.run_message()
        if not os.path.exists("alphafold"):
            subprocess.run(["git",
                    "clone",
                    "https://github.com/deepmind/alphafold.git"])
        module_dir = os.getcwd()
        if My_docker.runs == True and My_docker.installed == True:
            os.chdir(os.path.join(module_dir, "alphafold"))
            if not My_docker.is_image_built():
                subprocess.run(["docker",
                                "build",
                                "-f",
                                "docker/Dockerfile",
                                "-t",
                                "alphafold",
                                "."])  # "." indicates current directory as build context
            
            run_alphafold(os.path.join(module_dir, "temp", filename + ".fasta"), output_dir)
    os.chdir(module_dir)
        
folding("QMQLVQSGAEVKKPGSSVKVSCKASGGTFSSYAISWVRQAPGQGLEWMGGIIPIFGTANYAQKFQGRVTITADESTSTAYMELSSLRSEDTAVYYCARSSGYYGMDVWGQGTLVTVSS", r"C:\Users\nilsh\my_projects\ExpoSeq\my_experiments\test_true\alphafold_results")




import subprocess
import os
import shutil
import os
from ast import literal_eval
    


class Settings:
    def __init__(self):
        self.module_dir = os.path.abspath("")
        self.common_vars =  os.path.join(self.module_dir,"settings", "global_vars.txt")
        self.font_settings_path = os.path.join(self.module_dir, "settings", "font_settings.txt")
        self.legend_settings_path = os.path.join(self.module_dir,
                                        "settings",
                                        "legend_settings.txt")
        self.colorbar_settings_path = os.path.join(self.module_dir,
                                                    "settings",	
                                                    "colorbar.txt")
        
    def check_dirs(self):
        if not os.path.isdir(os.path.join(self.module_dir, "my_experiments")):
            print("create my_experiments directory")
            os.mkdir(os.path.join(self.module_dir, "my_experiments"))
        if not os.path.isdir(os.path.join(self.module_dir, "temp")):
            print("create temp directory")
            os.mkdir(os.path.join(self.module_dir, "temp"))
        if not os.path.isdir(os.path.join(self.module_dir, "settings")):
            print("create settings directory")
            os.mkdir(os.path.join(self.module_dir, "settings"))
            
    def create_global_vars(self):
        global_params = {'mixcr_path': '', 'last_experiment': '', 'api_gpt3': '', 'region_of_interest': '', 'RAM': '', 'clustalw_path': ''}
        with open(self.common_vars, "w") as f:
            f.write(str(global_params))
    
    def create_font_settings(self):
        
        font_settings = {'fontfamily': 'serif', 'fontsize': '18', 'fontstyle': 'normal', 'fontweight': 'bold'}
        with open(self.font_settings_path, "w") as f:
            f.write(str(font_settings))
            
    def create_legend_settings(self):
        legend_params = {'loc': 'center left', 'bbox_to_anchor': (1, 0.5), 'fontsize': 9, 'frameon': True, 'framealpha': 1, 'facecolor': 'white', 'mode': None, 'title_fontsize': 'small'}
        with open(self.legend_settings_path, "w") as f:
            f.write(str(legend_params))
            
    def create_colorbar_settings(self):
        colorbar_settings = {'cmap': 'inferno', 'orientation': 'vertical', 'spacing': 'proportional', 'extend': 'neither'}
        with open(self.colorbar_settings_path, "w") as f:
            f.write(str(colorbar_settings))
            
            
    def read_global_vars(self):
        if not os.path.isfile(self.common_vars):
            self.create_global_vars()
        with open(self.common_vars, "r") as f:
            global_params = f.read()
        global_params = literal_eval(global_params)
        return global_params
        
    def change_ram(self,ram):
        self.global_params["RAM"] = ram
        with open(self.common_vars, "w") as f:
            f.write(str(self.global_params))
    
    def change_global_vars(self, param, input):
        self.global_params[param] = input
        with open(self.common_vars, "w") as f:
            f.write(str(self.global_params))
            
    def change_mixcr_path(self, path):
        global_params = self.read_global_vars()
        self.global_params["mixcr_path"] = path
        with open(self.common_vars, "w") as f:
            f.write(str(self.global_params))
            

            
    def save_settings(self, path, params):
            with open(path, "w") as f:
                f.write(str(params))
    def remove_directories(directory_path):
    # Iterate over all items in the directory
        for item in os.listdir(directory_path):
            shutil.rmtree(os.path.join(directory_path, item))
            
    def read_font_settings(self):
        if not os.path.isfile(self.font_settings_path):
            self.create_font_settings()
        font_settings_path = os.path.join(self.module_dir, "settings", "font_settings.txt")
        assert os.path.isfile(font_settings_path), "The font settings file does not exist in the given filepath"
        with open(font_settings_path, "r") as f:
            font_settings = f.read()
        font_settings = literal_eval(font_settings)
        return font_settings
    
    def read_legend_settings(self):
        if not os.path.isfile(self.legend_settings_path):
            self.create_legend_settings()
        legend_settings_path = os.path.join(self.module_dir,
                                        "settings",
                                        "legend_settings.txt")
        assert os.path.isfile(legend_settings_path), "The legend settings file does not exist in the given filepath"
        with open(legend_settings_path, "r") as f:
            legend_settings = f.read()
        legend_settings = literal_eval(legend_settings)
        return legend_settings
    
    def read_colorbar_settings(self):
        if not os.path.isfile(self.colorbar_settings_path):
            self.create_colorbar_settings()
        colorbar_path = os.path.join(self.module_dir,
                                    "settings",
                                    "colorbar.txt")
        assert os.path.isfile(colorbar_path), "The colorbar file does not exist in the given filepath"
        with open(colorbar_path, "r") as f:
            colorbar_settings = f.read()
        colorbar_settings = literal_eval(colorbar_settings)
        return colorbar_settings
    
    def get_experiment_path(self, experiment):
        experiment_path = os.path.join(self.module_dir,
                                "my_experiments",
                                experiment,
                                "experiment_names.pickle")
        return experiment_path


    def reset_pipeline(self):
        global_params = {'mixcr_path': '', 'last_experiment': '', 'api_gpt3': '', 'region_of_interest': '', 'RAM': '', 'clustalw_path': ''}
        self.save_settings(self.global_params, global_params)
        self.save_settings(os.path.join(self.module_dir,"settings","save_settings.txt"), {'fname': '', 'format': 'png', 'dpi': 300})
        legend_params = {"loc": "upper right", #'upper left', 'upper right', 'lower left', 'lower right'
                    "bbox_to_anchor": (1, 1),
                    "ncols": 1,
                    "fontsize": "medium", #'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'
                    "frameon": True,
                    "framealpha": 1,
                    "facecolor": "white", #background color
                    "mode": None,
                    "title_fontsize": 'small',
                    }
        self.save_settings(os.path.join(self.module_dir,"settings","legend_settings.txt"), legend_params)
        self.save_settings(os.path.join(self.module_dir,"settings","font_settings.txt"), {"fontfamily": "serif","fontsize": "14","fontstyle": "normal","fontweight": "bold"})
        self.save_settings(os.path.join(self.module_dir,"settings","colorbar.txt"), {"cmap": "inferno","orientation": "vertical","spacing": "proportional","extend": "neither"})
        

    def reinstall_pipeline(self,):
        subprocess.run(["pip","-f", "uninstall", "ExpoSeq"])
        subprocess.run(["pip","-f", "install", "ExpoSeq"])
        delete = input("You will now delete all your analyzed experiments. Press enter to cancel this operation or type 'delete' to continue.")
        if delete == "delete":
            self.remove_directories(os.path.join(self.module_dir, "my_experiments"))
            self.remove_directories(os.path.join(self.module_dir, "temp"))
            
            


import pandas as pd
def find_seq_matches(sequencing_report,region_of_interest, protein, specific_experiments):
    if specific_experiments != False:
        sequencing_report = sequencing_report[sequencing_report['Experiment'].isin(specific_experiments)]
    else:
        pass
    unique_experiments = sequencing_report["Experiment"].unique()
    if protein == True:
        strand_column = region_of_interest
    else:
        strand_column =  region_of_interest
    unique_sequences = pd.DataFrame(sequencing_report[strand_column].unique())
    unique_sequences.rename(columns={0: strand_column},
                            inplace=True)

    for i, experiment in enumerate(unique_experiments):
        local_data = sequencing_report[sequencing_report["Experiment"] == experiment][strand_column].unique()
        local_data = pd.DataFrame(local_data, columns=[strand_column])
        unique_sequences = pd.merge(unique_sequences, local_data,
                                    on=strand_column,
                                    how='left',
                                    indicator=True)
        unique_sequences[experiment] = unique_sequences["_merge"].eq("both")
        unique_sequences = unique_sequences.drop("_merge", axis=1)
    unique_sequences.drop(strand_column,
                          inplace=True,
                          axis=1)
    unique_sequences = unique_sequences.astype("int")
    heatmap_axis = len(unique_experiments)
    return heatmap_axis, unique_sequences, unique_experiments


data_list = [
    "10x-5gex-cdr3",
    "10x-5gex-full-length",
    "10x-vdj-bcr",
    "10x-vdj-bcr-full-length",
    "10x-vdj-tcr",
    "10x-vdj-tcr-full-length",
    "abhelix-human-bcr-cdr3",
    "abhelix-human-bcr-full-length",
    "abhelix-human-tcr-cdr3",
    "abhelix-human-tcr-full-length",
    "ampliseq-tcrb-plus-cdr3",
    "ampliseq-tcrb-plus-full-length",
    "ampliseq-tcrb-sr-cdr3",
    "bd-rhapsody-human-bcr-full-length",
    "bd-rhapsody-human-bcr-v1",
    "bd-rhapsody-human-bcr-v2",
    "bd-rhapsody-human-tcr-full-length",
    "bd-rhapsody-human-tcr-v1",
    "bd-rhapsody-human-tcr-v2",
    "bd-rhapsody-mouse-bcr-full-length",
    "bd-rhapsody-mouse-bcr-v1",
    "bd-rhapsody-mouse-bcr-v2",
    "bd-rhapsody-mouse-tcr-full-length",
    "bd-rhapsody-mouse-tcr-v1",
    "bd-rhapsody-mouse-tcr-v2",
    "biomed2-human-bcr-cdr3",
    "biomed2-human-bcr-full-length",
    "cellecta-air-human",
    "irepertoire-human-dna-igh-lr",
    "irepertoire-human-dna-igh-sr",
    "irepertoire-human-dna-trb-lr",
    "irepertoire-human-dna-trb-sr",
    "irepertoire-human-rna-bcr-lr",
    "irepertoire-human-rna-bcr-sr",
    "irepertoire-human-rna-repseq-plus",
    "irepertoire-human-rna-tcr-lr",
    "irepertoire-human-rna-tcr-sr",
    "irepertoire-mouse-rna-bcr-lr",
    "irepertoire-mouse-rna-bcr-sr",
    "irepertoire-mouse-rna-tcr-lr",
    "irepertoire-mouse-rna-tcr-sr",
    "mikelov-et-al-2021",
    "milab-human-bcr-multiplex-cdr3",
    "milab-human-bcr-multiplex-full-length",
    "milab-human-tcr-dna-multiplex-cdr3",
    "milab-human-tcr-rna-multiplex-cdr3",
    "milab-human-tcr-rna-race-cdr3",
    "milab-human-tcr-rna-race-full-length",
    "milab-mouse-tcr-rna-race-cdr3",
    "milab-mouse-tcr-rna-race-full-length",
    "nebnext-human-bcr-base",
    "nebnext-human-bcr-cdr3",
    "nebnext-human-bcr-full-length",
    "nebnext-human-tcr-base",
    "nebnext-human-tcr-cdr3",
    "nebnext-human-tcr-full-length",
    "nebnext-mouse-bcr-base",
    "nebnext-mouse-bcr-cdr3",
    "nebnext-mouse-bcr-full-length",
    "nebnext-mouse-tcr-base",
    "nebnext-mouse-tcr-cdr3",
    "nebnext-mouse-tcr-full-length",
    "oncomine-human-bcr-igh-lr-full-length",
    "oncomine-human-bcr-igh-sr-cdr3",
    "oncomine-human-bcr-ihg-lr-cdr3",
    "oncomine-human-bcr-pan-clonality-cdr3",
    "oncomine-human-igh-fr1-j",
    "oncomine-human-igh-fr2-j",
    "oncomine-human-igh-fr3-j",
    "oncomine-human-igh-leader-j",
    "oncomine-human-tcr-pan-clonality-cdr3",
    "oncomine-human-tcrb-lr-cdr3",
    "oncomine-human-tcrb-lr-full-length",
    "oncomine-human-tcrb-sr-cdr3",
    "parse-bio-vdj-3gex",
    "qiaseq-human-tcr-cdr3",
    "qiaseq-human-tcr-full-length",
    "qiaseq-mouse-tcr-cdr3",
    "qiaseq-mouse-tcr-full-length",
    "singleron-2.0.1-vdj-cdr3",
    "split-seq-vdj-3gex",
    "takara-human-bcr-cdr3",
    "takara-human-bcr-full-length",
    "takara-human-tcr-V1-cdr3",
    "takara-human-tcr-V1-full-length",
    "takara-human-tcr-V2-cdr3",
    "takara-human-tcr-V2-full-length",
    "takara-mouse-bcr-cdr3",
    "takara-mouse-bcr-full-length",
    "takara-mouse-tcr-cdr3",
    "takara-mouse-tcr-full-length",
    "vergani-et-al-2017-cdr3",
    "vergani-et-al-2017-full-length"
]



from ExpoSeq.tidy_data.tidy_rel_sequence_abun import cleaning
import numpy as np
from ExpoSeq.plots.layout_finder import best_layout

def relative_sequence_abundance(ax, sequencing_report, samples,max_levenshtein_distance,length_filter,batch,region_string, font_settings, legend_settings):
    all_samples = cleaning(sequencing_report,
                            max_levenshtein_distance,
                            samples,
                            length_filter,
                            region_string,
                            batch)
    x_vals = all_samples.index.to_list()
    bar_width = 0.25
    all_samples_columns = all_samples.columns.to_list()
    for i in range(all_samples.shape[1]):
        y = all_samples.iloc[:, i]
        x_pos = [j + i * bar_width for j in range(len(x_vals))]

        ax.bar(x_pos,
               y,
               width=0.25,
               align='center',
               label=all_samples_columns[i])
    ax.set_xticks(range(len(x_vals)),
                  x_vals,
                  rotation = 45,
                  ha = "right")
    ax.set_ylabel("Frequency in samples with Levenshtein Distance " + str(max_levenshtein_distance),
                  **font_settings)
    ax.set_xlabel("Sequences",
                  **font_settings)
    ax.legend(**legend_settings)
    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 22
    ax.set_title("Sequence Abundance of given samples",
                 pad = 12,
                 **font_settings)
    font_settings["fontsize"] = original_fontsize
    
def relative_sequence_abundance_all(fig, sequencing_report, samples, font_settings, region_string):
    if samples == "all":
        unique_experiments = sequencing_report["Experiment"].unique()
        unique_experiments = np.sort(unique_experiments)
    else:
        sequencing_report = sequencing_report[sequencing_report['Experiment'].isin(samples)]
        unique_experiments = sequencing_report["Experiment"].unique()
        unique_experiments = np.sort(unique_experiments)
        
    Tot = unique_experiments.shape[0]
    Rows, Cols = best_layout(Tot)
    Position = range(1, Tot + 1)
    n = 0
    
    for experiment in unique_experiments:
        batch = sequencing_report[sequencing_report["Experiment"] == experiment]
        fraction = batch["cloneFraction"]
        clones = batch["cloneId"]
        ax = fig.add_subplot(Rows, Cols, Position[n])
        
        ax.bar(clones, fraction)  # Or whatever you want in the subplot
       # ax.set_xticks(range(0, max_length + 1, 1), range(0, max_length + 1, 1))
        ax.title.set_text(experiment)
        ax.title.set_size(10)

        adapted_fontsize = 10 - int(Cols) + 2
        font_settings["fontsize"] = adapted_fontsize
        ax.set_ylabel("Clones",
                      **font_settings)  # Y label
        ax.set_xlabel('Clone Fraction',
                      **font_settings)  # X label

        ax.set_ylim(0, np.max(np.array(fraction) + np.max(np.array(fraction) * 0.1)))
        n += 1
    fig.suptitle("Clone Fraction of given Samples")




import pandas as pd
import os

try:
    import tkinter as tk
    from tkinter import filedialog
except:
    pass

def collect_binding_data(binding_data = None):
    if binding_data is None:
        binding_data = pd.DataFrame([])
    else:
        pass
    second_prompt = False
    while True:
        # prompt the user to add a file
        if not second_prompt:
            print("You can either add an excel sheet or a csv file which contains the binding data.\nNote: the first column must contain the CDR3 sequences and its column name has to be aaSeqCDR3.")
        else:
            print("Please modify the file and choose it again.")
        try:
            binding_file = filedialog.askopenfilename()
        except:
            while True:
                binding_file = input("copy and paste the path to your binding report")
                if os.path.isfile(os.path.abspath(binding_file)):
                    break
                else:
                    print("Please enter a valid filepath. ")
        
        if binding_file.endswith(".xlsx") or binding_file.endswith(".csv") or binding_file.endswith(".tsv"):
            if binding_file.endswith(".xlsx"):
                binding_new = pd.read_excel(binding_file)
            elif binding_file.endswith(".csv"):
                binding_new = pd.read_csv(binding_file)
            elif binding_file.endswith(".tsv"):
                binding_new = pd.read_table(binding_file)
            if binding_new.columns.to_list()[0] == "aaSeqCDR3":
                second_prompt = False
                pass
            else:
                second_prompt = True
                print("Please chechk if your file follows the following requirements:\n1. The first column must contain the CDR3 sequences and its column name has to be aaSeqCDR3.\n2. Check that you separate your data by comma or tab.\n3. Your file needs to end with .xlsx, .csv or .tsv")
        else:
            print("Please enter a valid filepath to a csv or xlsx file")
            second_prompt = True

        if not second_prompt:
            binding_data = pd.concat([binding_data, binding_new])
            response = input("Do you want to continue adding files? (Y/n) ")
            if response.lower() == "n":
                break
            print("The first five rows of your binding data look like this:")
            print(binding_data.head(5))
    return binding_data


try: 
    import tkinter as tk
    from tkinter import simpledialog
except:
    pass

class SequenceManipulator:
    def __init__(self, sequence):
        self.root = tk.Tk()
        self.root.title('Sequence Manipulator')

        # Set the root window's height and width
        self.root.geometry("800x600")

        # Create a frame for canvas and scrollbar
        frame = tk.Frame(self.root, height=150)  # Define the frame's height
        frame.pack(fill="x", padx=20, pady=20)

        self.canvas = tk.Canvas(frame, height=150)  # Define canvas height
        self.canvas.pack(fill="x", expand=True, side="top")

        scrollbar = tk.Scrollbar(frame, command=self.canvas.xview, orient="horizontal")
        scrollbar.pack(side="bottom", fill="x")

        # Create the canvas frame and attach to the canvas
        self.canvas_frame = tk.Frame(self.canvas)
        self.canvas.create_window((0, 0), window=self.canvas_frame, anchor="nw")

        self.sequence = tk.StringVar()
        self.sequence.set(sequence)

        spacing = 20  # consistent spacing
        for idx, char in enumerate(sequence):
            self.canvas.create_text(10 + idx * spacing, 10, anchor="nw", text=char, font=('Courier', 10))

        # Add positions below the sequence
        positions = [str(i) if i % 5 == 0 else '' for i in range(1, len(sequence) + 1)]
        for idx, position in enumerate(positions):
            pos_width = (idx + 1) * spacing
            self.canvas.create_text(10 + idx * spacing, 30, anchor="nw", text=position, font=('Courier', 10))

        # Update scroll region after items are added to canvas
        self.canvas.config(xscrollcommand=scrollbar.set)
        self.canvas.update_idletasks()  # ensure the canvas items are created
        self.canvas.config(scrollregion=self.canvas.bbox("all"))

        # Frame for buttons
        btn_frame = tk.Frame(self.root)
        btn_frame.pack(fill="both", pady=20, expand=True)

        btn_insert = tk.Button(btn_frame, text="Insert AA", command=self.insert_amino_acid)
        btn_insert.pack(side="left", fill="both", expand=True, padx=5, pady=10)

        btn_delete = tk.Button(btn_frame, text="Delete AA", command=self.delete_amino_acid)
        btn_delete.pack(side="left", fill="both", expand=True, padx=5, pady=10)

        btn_replace = tk.Button(btn_frame, text="Replace AA", command=self.replace_amino_acid)
        btn_replace.pack(side="left", fill="both", expand=True, padx=5, pady=10)

    def insert_amino_acid(self):
        position = simpledialog.askinteger("Input", "At which position?", parent=self.root, minvalue=1, maxvalue=len(self.sequence.get())+1)
        if position is None: return
        aa = simpledialog.askstring("Input", "Which amino acid to insert?", parent=self.root)
        if not aa: return
        current_seq = self.sequence.get()
        new_seq = current_seq[:position-1] + aa + current_seq[position-1:]
        self.sequence.set(new_seq)

    def delete_amino_acid(self):
        position = simpledialog.askinteger("Input", "At which position?", parent=self.root, minvalue=1, maxvalue=len(self.sequence.get()))
        if position is None: return
        current_seq = self.sequence.get()
        new_seq = current_seq[:position-1] + current_seq[position:]
        self.sequence.set(new_seq)

    def replace_amino_acid(self):
        position = simpledialog.askinteger("Input", "At which position?", parent=self.root, minvalue=1, maxvalue=len(self.sequence.get()))
        if position is None: return
        aa = simpledialog.askstring("Input", "Replace with which amino acid?", parent=self.root)
        if not aa: return
        current_seq = self.sequence.get()
        new_seq = current_seq[:position-1] + aa + current_seq[position:]
        self.sequence.set(new_seq)

    def run(self):
        self.root.mainloop()




import pandas as pd
from ExpoSeq.tidy_data.read_extract_data import read_alignment_report



def collect_boxplot_data(report_filename, local_pattern):
    with open(report_filename) as f:
        alignment_report = f.readlines()
    data_extractor = read_alignment_report(alignment_report)
    digit = data_extractor.digit
    total_sequencing_reads = data_extractor.find_word_align_rep(word = "Total sequencing reads",
                                                                regex = digit) # digit in class
    total_sequencing_reads = int(total_sequencing_reads)
    successfully_aligned_reads = data_extractor.find_word_align_rep(word = "Successfully aligned reads",
                                                                    regex = digit)
    successfully_aligned_reads = int(successfully_aligned_reads)
    experiment = data_extractor.find_word_align_rep("Output file",
                                                    regex = local_pattern)
    data = {'Experiment': [experiment],
            'Aligned_Reads': [successfully_aligned_reads],
            'tot_sequenced_reads': [total_sequencing_reads]}
    alignment_frame = pd.DataFrame(data)
    return alignment_frame


from tkinter import TclError
import tkinter.font as tkfont
import tkinter as tk
from tkinter import simpledialog, messagebox
from tkinter.colorchooser import askcolor
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import logomaker
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
AMINO_ACID_COLORS = {
    "S": "green", "T": "green","Q": "green","N": "green",
    "A": "grey", "V": "grey", "L": "grey", "I": "grey", "F": "grey",  "W": "grey", "M": "grey", "Y": "grey",
    "C": "yellow","P": "yellow","G": "yellow",
    "D": "red", "E": "red",
    "H": "cyan","K": "blue","R": "blue",
    '-': 'black', '*': 'black','end': "black"
}
AMINO_ACID_COLORS = {
    "S": "#388E3C",  # Green 700
    "T": "#388E3C",  # Green 700
    "Q": "#7CB342",  # Light Green 600
    "N": "#7CB342",  # Light Green 600
    "A": "#616161",  # Grey 700
    "V": "#616161",  # Grey 700
    "L": "#424242",  # Grey 800
    "I": "#424242",  # Grey 800
    "F": "#212121",  # Grey 900
    "W": "#212121",  # Grey 900
    "M": "#455A64",  # Blue Grey 700
    "Y": "#455A64",  # Blue Grey 700
    "C": "#F57C00",  # Orange 700 (Instead of Yellow)
    "P": "#F57C00",  # Orange 700 (Instead of Yellow)
    "G": "#FFA000",  # Amber 700 (Brighter than Orange)
    "D": "#D32F2F",  # Red 700
    "E": "#C62828",  # Red 800
    "H": "#00838F",  # Cyan 700
    "K": "#1565C0",  # Blue 700
    "R": "#0D47A1",  # Blue 900
    '-': "#1B1B1B",  # Almost Black
    '*': "#1B1B1B",  # Almost Black
    'end': "#FFFFFF", # White,
    '.': "#1B1B1B",
    ':': "#1B1B1B"
}

def parse_clustal_format(filename):
    sequences = {}
    with open(filename, 'r') as file:
        lines = file.readlines()
        # Remove any initial non-sequence lines (e.g., metadata or blank lines)
        while lines and (lines[0].strip() == '' or lines[0].startswith('CLUSTAL') or lines[0].startswith(' ')):
            lines.pop(0)
        while lines:
            seq_block = {}
            while lines and lines[0].strip() != '':
                line = lines.pop(0).strip()
                if not line.startswith(" "):  # Avoiding lines starting with space which can be alignment indicators
                    parts = line.split()
                    seq_id = parts[0]
                    seq_segment = parts[1]
                    if seq_id not in seq_block:
                        seq_block[seq_id] = seq_segment
                    else:
                        seq_block[seq_id] += seq_segment
            # Update the master sequence dictionary
            for seq_id, seq_segment in seq_block.items():
                if seq_id not in sequences:
                    sequences[seq_id] = seq_segment
                else:
                    sequences[seq_id] += seq_segment
            # Skip any spacer lines between sequence blocks
            while lines and lines[0].strip() == '':
                lines.pop(0)
    return sequences



class SequenceManipulator:
    def __init__(self, sequence, spacing = 20):
        self.root = tk.Tk()
        self.root.title('Sequence Manipulator')

        # Set the root window's height and width
        self.root.geometry("800x600")

        # Create a frame for canvas and scrollbar
        frame = tk.Frame(self.root, height=150)  # Define the frame's height
        frame.pack(fill="x", padx=20, pady=20)

        self.canvas = tk.Canvas(frame, height=150)  # Define canvas height
        self.canvas.pack(fill="x", expand=True, side="top")

        scrollbar = tk.Scrollbar(frame, command=self.canvas.xview, orient="horizontal")
        scrollbar.pack(side="bottom", fill="x")

        # Create the canvas frame and attach to the canvas
        self.canvas_frame = tk.Frame(self.canvas)
        self.canvas.create_window((0, 0), window=self.canvas_frame, anchor="nw")

        self.sequence = tk.StringVar()
        self.sequence.set(sequence)

        self.spacing = spacing  # consistent spacing
        for idx, char in enumerate(sequence):
            self.canvas.create_text(10 + idx * self.spacing, 10, anchor="nw", text=char, font=('Courier', 10),
                                    tags="sequence_char")

        # Add positions below the sequence
        positions = [str(i) if i % 5 == 0 else '' for i in range(1, len(sequence) + 1)]
        for idx, position in enumerate(positions):
            pos_width = (idx + 1) * self.spacing
            self.canvas.create_text(10 + idx * self.spacing, 30, anchor="nw", text=position, font=('Courier', 10))
        self.indicator_position = 0
        self.clipboard = ""
        # Update scroll region after items are added to canvas
        self.canvas.config(xscrollcommand=scrollbar.set)
        self.canvas.update_idletasks()  # ensure the canvas items are created
        self.canvas.config(scrollregion=self.canvas.bbox("all"))
        self.canvas.bind("<Button-1>", self.on_canvas_click)
        self.clipboard = ""

        self.canvas.bind("<Button-1>", self.focus_canvas)

        self.canvas.bind("<Left>", self.move_indicator_left)
        self.canvas.bind("<Right>", self.move_indicator_right)
        self.canvas.bind("<Key>", self.on_key_press)

        self.canvas.bind("<BackSpace>", self.on_backspace)
        self.canvas.bind('<Control-v>', self.on_paste)
        self.canvas.bind('<Control-c>', self.on_copy)

        # If you're on macOS, also add:
        self.canvas.bind("<Command-c>", self.on_copy)
        self.canvas.bind("<Command-v>", self.on_paste)

        self.canvas.focus_set()

        # Frame for buttons
        btn_frame = tk.Frame(self.root)
        btn_frame.pack(fill="both", pady=20, expand=True)

        btn_insert = tk.Button(btn_frame, text="Insert AA", command=self.insert_amino_acid)
        btn_insert.pack(side="left", fill="both", expand=True, padx=5, pady=10)

        btn_delete = tk.Button(btn_frame, text="Delete AA", command=self.delete_amino_acid)
        btn_delete.pack(side="left", fill="both", expand=True, padx=5, pady=10)

        btn_replace = tk.Button(btn_frame, text="Replace AA", command=self.replace_amino_acid)
        btn_replace.pack(side="left", fill="both", expand=True, padx=5, pady=10)

        btn_insert_region = tk.Button(btn_frame, text="Insert Region", command=self.insert_region)
        btn_insert_region.pack(side="left", fill="both", expand=True, padx=5, pady=10)

        btn_replace_region = tk.Button(btn_frame, text="Replace Region", command=self.replace_region)
        btn_replace_region.pack(side="left", fill="both", expand=True, padx=5, pady=10)

    def on_copy(self, event):
        idx = (self.indicator_position - 10) // self.spacing
        if 0 <= idx < len(self.sequence.get()):
            char_to_copy = self.sequence.get()[idx]
            self.clipboard_clear()
            self.clipboard_append(char_to_copy)
            print(f"Copied: {char_to_copy}")
        else:
            print("Nothing to copy")

    def on_paste(self, event):
        try:
            clipboard_text = self.text_widget.clipboard_get()
            # Now, you can process or use the clipboard_text as you wish
            # For example, if you want to insert it into the Text widget:
            self.text_widget.insert(tk.END, clipboard_text)
        except:
            print("error")

    #     except TclError:  # No data in clipboard
          #  print("Clipboard is empty")

    def focus_canvas(self, event):
        self.canvas.focus_set()
    def on_key_press(self, event):
        char = event.char.upper()
        if char in 'ACDEFGHIKLMNPQRSTVWY':  # This allows any character, modify as needed.
            idx = (self.indicator_position - 10) // self.spacing
            current_seq = self.sequence.get()
            updated_seq = current_seq[:idx] + char + current_seq[idx:]
            self.sequence.set(updated_seq)
            self.move_indicator_right(None)  # Move the caret to the right after insertion
            self.draw_sequence()

    def on_backspace(self, event):
        idx = (self.indicator_position - 10) // self.spacing
        if idx >= 0:  # This will allow the first character to be deleted
            current_seq = self.sequence.get()
            updated_seq = current_seq[:idx] + current_seq[idx + 1:]
            self.sequence.set(updated_seq)
            self.move_indicator_left(None)
            self.draw_sequence()

    def insert_sequence_at_position(self, idx):
        new_sequence = self.entry.get().upper()  # Get the content of the entry widget
        amino_acids = 'ACDEFGHIKLMNPQRSTVWY'  # 20 standard amino acids
        if all(char in amino_acids for char in new_sequence):
            # Check if the entered characters are valid nucleotides
            current_seq = self.sequence.get()
            updated_seq = current_seq[:idx] + new_sequence + current_seq[idx:]
            self.sequence.set(updated_seq)
            self.draw_sequence()
            self.entry.destroy()  # Destroy the entry widget after inserting
        else:
            messagebox.showwarning("Invalid Input", "Please enter a valid nucleotide sequence (A, C, G, T).")

    def move_indicator_left(self, event):
        # Decrease the indicator position by the spacing only if it won't go beyond the starting position
        if self.indicator_position > 10:
            self.indicator_position -= self.spacing
            self.draw_position_indicator()

    def move_indicator_right(self, event):
        # Increase the indicator position by the spacing
        if self.indicator_position < (len(self.sequence.get()) * self.spacing):
            self.indicator_position += self.spacing
            self.draw_position_indicator()

    def draw_position_indicator(self):
        self.canvas.delete("position_indicator")
        self.canvas.create_line(self.indicator_position, 0, self.indicator_position, 50, tags="position_indicator",
                                fill="black")

    def on_key_press(self, event):
        char = event.char.upper()
        if char in 'ACDEFGHIKLMNPQRSTVWY':  # If it's an amino acid
            idx = (self.indicator_position - 10) // self.spacing
            current_seq = self.sequence.get()
            updated_seq = current_seq[:idx] + char + current_seq[idx:]
            self.sequence.set(updated_seq)
            self.move_indicator_right(None)
            self.draw_sequence()
        elif event.keysym == "BackSpace":
            self.on_backspace(event)

    def on_canvas_click(self, event):
        # Remove any existing vertical lines (with the tag "position_indicator")
        self.canvas.delete("position_indicator")

        # Calculate the closest position index based on the click
        idx = round((event.x - 10) / self.spacing)
        if idx >= len(self.sequence.get()):  # if clicked beyond the last character
            idx = len(self.sequence.get()) - 1
        self.indicator_position = 10 + idx * self.spacing

        # Draw the position indicator at the calculated position
        self.draw_position_indicator()

        # Calculate approximate sequence index from x position
        idx = (event.x - 10) // self.spacing  # Assuming 10 is the starting x-coordinate, adjust if different.

        # Position the Entry widget at the clicked position
        self.entry = tk.Entry(self.canvas_frame, width=5)
        self.entry.place(x=event.x, y=10)
        self.entry.focus_set()  # Set focus to Entry widget

        # Bind events to the entry
        self.entry.bind("<Return>", lambda e: self.insert_sequence_at_position(idx))
        self.entry.bind("<Escape>", lambda e: self.entry.destroy())  # Destroy entry widget if escape key pressed

    def draw_positions(self):
        self.canvas.delete("position_number")  # Clear previous positions
        positions = [str(i) if i % 5 == 0 else '' for i in range(1, len(self.sequence.get()) + 1)]
        for idx, position in enumerate(positions):
            self.canvas.create_text(10 + idx * self.spacing, 30, anchor="nw", text=position, font=('Courier', 10),
                                    tags="position_number")
    def draw_sequence(self):
        self.canvas.delete("all")
        self.draw_positions()
  # Clear previous sequence

        sequence = self.sequence.get()
        for idx, char in enumerate(sequence):
            self.canvas.create_text(10 + idx * self.spacing, 10, anchor="nw", text=char, font=('Courier', 10),
                                    tags="sequence_char")
        self.draw_position_indicator()

    def insert_amino_acid(self):
        position = simpledialog.askinteger("Input", "At which position?", parent=self.root, minvalue=1, maxvalue=len(self.sequence.get())+1)
        if position is None: return
        aa = simpledialog.askstring("Input", "Which amino acid to insert?", parent=self.root)
        if not aa: return
        current_seq = self.sequence.get()
        new_seq = current_seq[:position-1] + aa + current_seq[position-1:]
        self.sequence.set(new_seq)
        self.draw_sequence()
    def delete_amino_acid(self):
        position = simpledialog.askinteger("Input", "At which position?", parent=self.root, minvalue=1, maxvalue=len(self.sequence.get()))
        if position is None: return
        current_seq = self.sequence.get()
        new_seq = current_seq[:position-1] + current_seq[position:]
        self.sequence.set(new_seq)
        self.draw_sequence()

    def replace_amino_acid(self):
        position = simpledialog.askinteger("Input", "At which position?", parent=self.root, minvalue=1, maxvalue=len(self.sequence.get()))
        if position is None: return
        aa = simpledialog.askstring("Input", "Replace with which amino acid?", parent=self.root)
        if not aa: return
        current_seq = self.sequence.get()
        new_seq = current_seq[:position-1] + aa + current_seq[position:]
        self.sequence.set(new_seq)
        self.draw_sequence()

    def insert_region(self):
        start_position = simpledialog.askinteger("Input", "Starting position?", parent=self.root, minvalue=1, maxvalue=len(self.sequence.get())+1)
        end_position = simpledialog.askinteger("Input", "Ending position?", parent=self.root, minvalue=start_position, maxvalue=len(self.sequence.get())+1)
        if start_position is None or end_position is None: return
        region = simpledialog.askstring("Input", "Which region to insert?", parent=self.root)
        if not region: return
        current_seq = self.sequence.get()
        new_seq = current_seq[:start_position-1] + region + current_seq[end_position-1:]
        self.sequence.set(new_seq)
        self.draw_sequence()

    def replace_region(self):
        start_position = simpledialog.askinteger("Input", "Starting position of region to replace?", parent=self.root, minvalue=1, maxvalue=len(self.sequence.get())+1)
        end_position = simpledialog.askinteger("Input", "Ending position of region to replace?", parent=self.root, minvalue=start_position, maxvalue=len(self.sequence.get())+1)
        if start_position is None or end_position is None: return
        region = simpledialog.askstring("Input", "Replace with which region?", parent=self.root)
        if not region: return
        current_seq = self.sequence.get()
        new_seq = current_seq[:start_position-1] + region + current_seq[end_position:]
        self.sequence.set(new_seq)
        self.draw_sequence()

    def run(self):

        self.root.mainloop()

import re

def extract_number(seq_id):
    match = re.search(r'(\d+)', seq_id)
    return int(match.group(1)) if match else 999999  # Large default number


class RegionVisualizer:
    def __init__(self, parent, sequence_length, regions, char_width, x_offset, canvas_height=40):
        self.canvas_width = char_width * sequence_length
        self.canvas = tk.Canvas(parent, width=self.canvas_width, height=canvas_height)
        self.canvas.pack(fill="x", expand=True)
        self.canvas.config(scrollregion=(0, 0, self.canvas_width, canvas_height))

        for region in regions:
            start_pos = region['start']
            end_pos = region['end']
            color = region['color']
            label = region['label']

            x_start = x_offset + start_pos * char_width
            x_end = x_offset + end_pos * char_width

            self.canvas.create_line(x_start, canvas_height / 2, x_end, canvas_height / 2, fill=color, width=5)
            self.canvas.create_text((x_start + x_end) / 2, canvas_height / 4, text=label, anchor="center", font=('Courier', 8))

    def xview_moveto(self, fraction):
        self.canvas.xview_moveto(fraction)


class LimitedScrollText(tk.Text):
    def __init__(self, master=None, **kwargs):
        super().__init__(master, wrap=tk.NONE, **kwargs)
        self.bind("<Configure>", self.calculate_scroll_width)
        self.max_x_scroll = kwargs.pop('max_x_scroll', 1.0)

    def calculate_scroll_width(self, event=None):
        # Determine the maximum width of content in the widget
        content = self.get(1.0, tk.END).splitlines()
        longest_line = max(content, key=len, default="")

        # Measure the pixel width of the longest line
        font_name = self.cget('font')
        font = tkfont.Font(font=font_name)
        max_pixel_width = font.measure(longest_line)

        # Limit the xscrollcommand to the measured width
        self.xscrollcommand = lambda f, l: self.limit_xview_scroll(f, l, max_pixel_width)

    def limited_xview(self, *args):
        if args[0] == 'moveto':
            fraction = float(args[1])
            if 0 <= fraction <= self.max_x_scroll:
                self.xview(*args)
        else:
            # The scroll "units" scenario
            shift = int(args[1])
            current_x = self.xview()[0]  # Current x fraction
            target_fraction = current_x + (shift * self.unit_fraction)
            if 0 <= target_fraction <= self.max_x_scroll:
                self.xview(*args)
    def limit_xview_scroll(self, first, last, max_pixel_width):
        if float(last) <= 1.0:
            # Allow normal scrolling until the end of content
            self.xview_moveto(float(first))
        else:
            # Calculate the over-scroll distance in terms of fraction
            excess_fraction = max_pixel_width * (float(last) - 1.0) / (self.winfo_width() * float(last))
            # Limit the maximum scrolling position
            self.xview_moveto(float(first) - excess_fraction)

    # Override the default xview method to restrict its behavior
    def xview(self, *args):
        if args[0] == "moveto":
            fraction = float(args[1])
            self.xview_moveto(fraction)
        elif args[0] == "scroll":
            self.xview_scroll(int(args[1]), args[2])


class ExtendedSequenceManipulator(SequenceManipulator):
    def plot_amino_acid_composition(self):
        # Calculate amino acid composition for all sequences
        all_seqs = list(self.sequences.values()) # Change here
        print(all_seqs)
        unique_aas = list(set(all_seqs))
        chosen_sequence_length = len(all_seqs[0])
        aminoacids = "ACDEFGHIKLMNPQRSTVWY"
        compDict = {aa: chosen_sequence_length * [0] for aa in aminoacids}
        sequences = all_seqs
        for seq in sequences:
            for aa_position in range(len(seq)):
                aminoacid = seq[aa_position]
                if aminoacid in ["*", "-", "", ".", ":"]:
                    pass
                else:
                    compDict[aminoacid][aa_position] += 1
        aa_distribution = pd.DataFrame.from_dict(compDict)
        # aa_distribution = aa_distribution.divide(aa_distribution.shape[1])
        aa_distribution = aa_distribution.divide(len(all_seqs), axis=0)



        # Create the bar plot
        # Create the bar plot
        fig_width = chosen_sequence_length * 0.05  # adjust 0.05 as needed
  # dynamically set width based on sequence length
        fig, ax = plt.subplots(figsize=(fig_width, 4))
        logo_plot = logomaker.Logo(aa_distribution,
                                   shade_below=.5,
                                   fade_below=.5,
                                   font_name='Arial Rounded MT Bold',
                                   color_scheme="skylign_protein",
                                   show_spines=False,
                                   ax=ax,
                                   figsize= (fig_width, 4)
                                   )
        logo_plot.style_xticks(anchor=1,
                               spacing=1,
                               rotation=0)
        labels_true = list(range(0, chosen_sequence_length))
        numbers_true = list(range(1, chosen_sequence_length + 1))
        step = 5  # display every 5th label
        plt.xticks(labels_true[::step], numbers_true[::step])

        plot_frame = tk.Frame(self.root)
        plot_frame.pack(side=tk.TOP, fill=tk.BOTH) #expand 1  # pack it so that it fills available space

        # Embed the plot in the plot frame
        canvas = FigureCanvasTkAgg(fig, master=plot_frame)

        canvas_widget = canvas.get_tk_widget()


        canvas_widget.pack(side=tk.TOP, fill=tk.BOTH, expand=1)


        canvas.draw()  # Force drawing
    def __init__(self, sequence, fasta_file, master=None):
        super().__init__(sequence)

        # Regions and visual parameters
        regions = [
            {'start': 10, 'end': 50, 'color': 'red', 'label': 'Region A'},
            {'start': 55, 'end': 80, 'color': 'blue', 'label': 'Region B'}
        ]
        x_offset = 10
        self.char_width = 20

        # RegionVisualizer
       # self.region_visualizer = RegionVisualizer(self.root, len(sequence), regions, char_width, x_offset)

        # Parsing and sorting sequences
        self.sequences = parse_clustal_format(fasta_file)
        sorted_seq_ids = sorted(self.sequences.keys(), key=extract_number)
        self.sequences = {seq_id: self.sequences[seq_id] for seq_id in sorted_seq_ids}

        self.max_length = max(map(len, self.sequences.values()))

        # Multi-Sequence Alignment Display
        msa_frame = tk.Frame(self.root, height=400)
        msa_frame.pack(fill="both", expand=True)

        self.msa_display = LimitedScrollText(msa_frame, font=('Courier', 10))


        vertical_scroll = tk.Scrollbar(msa_frame, command=self.msa_display.yview, orient="vertical")
        horizontal_scroll = tk.Scrollbar(msa_frame, command=self.msa_display.limited_xview, orient="horizontal")

        # Configurations
        self.msa_display.config(yscrollcommand=vertical_scroll.set)
        horizontal_scroll.config(command=self.msa_display.xview)
        self.msa_display.config(xscrollcommand=horizontal_scroll.set)


        # Pack components
        horizontal_scroll.pack(side="bottom", fill="x")
        vertical_scroll.pack(side="right", fill="y")
        self.msa_display.pack(side="left", fill="both", expand=True)

        # Populate the MSA display with sequences
        # Populate the MSA display with sequences
        for seq_id, seq in self.sequences.items():
            formatted_seq = ' '.join([char for char in seq]).ljust(self.char_width * self.max_length)
            self.msa_display.insert("end", f">{seq_id}\n", "header")

            for char in seq:
                tag_name = f"{AMINO_ACID_COLORS[char]}_tag"
                self.msa_display.insert("end", char, tag_name)
                self.msa_display.insert("end", " ")
            self.msa_display.insert("end", "\n")
            spacing = 2
            # Insert positions below the sequence
            positions = [str(i) if i % 5 == 0 else '' for i in range(1, len(seq) + 1)]
            for idx, position in enumerate(positions):
                self.msa_display.insert("end", position.ljust(spacing))  # ljust will ensure each position takes up 'spacing' spaces.

            self.msa_display.insert("end", "\n\n")
        # Apply tag configurations
        for color in AMINO_ACID_COLORS.values():
            tag_name = f"{color}_tag"
            self.msa_display.tag_configure(tag_name, foreground=color)
        self.msa_display.tag_configure("header", font=("Courier", 10, "bold"))

        self.change_color_button = tk.Button(self.root, text="Change Color", command=self.on_change_color_button_pressed)
        self.change_color_button.pack(pady=5)
        self.zoom_scale = tk.Scale(self.root, from_=5, to=20, orient=tk.HORIZONTAL, label="Zoom", command=self.on_zoom)
        self.zoom_scale.set(10)  # set initial value
        self.zoom_scale.pack(pady=5)
       # self.plot_amino_acid_composition()
    # You can use this method to sync other visual elements with the scroll of the MSA.
    # For now, it's just a placeholder that only interacts with the MSA's own scroll.
    def on_zoom(self, value):
        font_size = int(value)
        self.msa_display.config(font=('Courier', font_size))

    def sync_with_msa_scroll(self, *args):
        fraction = float(args[0])
        self.region_visualizer.xview_moveto(fraction)
        self.msa_display.xview_moveto(fraction)

    def refresh_display(self):
        """Refresh the sequences displayed in the Text widget."""

        # Clear the existing content
        self.msa_display.delete(1.0, tk.END)

        # Insert the updated sequences
        for seq_id, seq in self.sequences.items():
            formatted_seq = ' '.join([char for char in seq]).ljust(self.char_width * self.max_length)
            self.msa_display.insert("end", f">{seq_id}\n", "header")

            for char in seq:
                tag_name = f"{AMINO_ACID_COLORS[char]}_tag"
                self.msa_display.insert("end", char, tag_name)
                self.msa_display.insert("end", " ")
            spacing = 2
            # Insert positions below the sequence
            positions = [str(i) if i % 5 == 0 else '' for i in range(1, len(seq) + 1)]
            for idx, position in enumerate(positions):
                self.msa_display.insert("end", position.ljust(spacing))  # ljust will ensure each position takes up 'spacing' spaces.

            self.msa_display.insert("end", "\n\n")

        for color in AMINO_ACID_COLORS.values():
            tag_name = f"{color}_tag"
            self.msa_display.tag_configure(tag_name, foreground=color)
        self.msa_display.tag_configure("header", font=("Courier", 10, "bold"))
        #self.draw_sequence()

    def change_color(self, amino_acid):
        """
        Allow the user to select a color for the provided amino acid and update the display.
        """
        color = askcolor()[1]  # Ask the user to pick a color
        if color:  # Check if a color was selected
            AMINO_ACID_COLORS[amino_acid] = color
            tag_name = f"{color}_tag"
            self.msa_display.tag_configure(tag_name, foreground=color)
            # You might also want to redraw or refresh your sequence display after this
        self.refresh_display()

    def on_change_color_button_pressed(self):
        amino_acid = simpledialog.askstring("Input", "Enter the amino acid you want to change the color for:")
        if amino_acid and amino_acid in AMINO_ACID_COLORS:
            self.change_color(amino_acid)


